from credsweeper.scanner.scan_type.scan_type import ScanType  # isort:skip

from credsweeper.scanner.scan_type.multi_pattern import MultiPattern
from credsweeper.scanner.scan_type.pem_key_pattern import PemKeyPattern
from credsweeper.scanner.scan_type.single_pattern import SinglePattern



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueNotAllowedPatternCheck(Filter):
    """Check that secret doesn't open or closes brackets or a new line."""

    NOT_ALLOWED = [r"[<>\[\]{}]\s+", r"\\u00(26|3c)gt;?(\s|\\+[nrt])?", r"^\s*\\", r"^\s*\\n\s*"]
    NOT_ALLOWED_PATTERN = re.compile(  #
        f"{Util.get_regex_combine_or(NOT_ALLOWED)}$",  #
        flags=re.IGNORECASE)

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if not line_data.is_well_quoted_value and self.NOT_ALLOWED_PATTERN.search(line_data.value):
            return True
        return False



import re

from credsweeper.common import static_keyword_checklist
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueCamelCaseCheck(Filter):
    """Check that candidate is not written in camel case."""

    CAMEL_CASE = ["^([a-z]+([A-Z][a-z]+)+)$", "^([A-Z][a-z]+([A-Z][a-z]+)+)$"]
    CAMEL_CASE_PATTERN = re.compile(Util.get_regex_combine_or(CAMEL_CASE))

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.is_well_quoted_value:
            return False
        if self.CAMEL_CASE_PATTERN.match(line_data.value):
            return static_keyword_checklist.check_morphemes(line_data.value.lower(), 1)

        return False



from typing import List, Any

import numpy as np

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class RuleName(WordIn):
    """Categorical feature that corresponds to rule name.

    Parameters:
        rule_names: rule name labels

    """

    def __init__(self, rule_names: List[str]) -> None:
        super().__init__(rule_names)

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        candidate_rule_set = set(x.rule_name for x in candidates)
        return self.word_in_set(candidate_rule_set)

    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError



import logging
import logging.config
from pathlib import Path
from typing import Optional

from credsweeper.app import APP_PATH
from credsweeper.utils import Util


class Logger:
    """Class that used to configure logging in CredSweeper."""

    SILENCE = 60

    LEVELS = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "FATAL": logging.CRITICAL,
        "CRITICAL": logging.CRITICAL,
        "SILENCE": SILENCE
    }

    @staticmethod
    def init_logging(log_level: str, file_path: Optional[str] = None) -> None:
        """Init logger.

        Init logging with configuration from file 'credsweeper_path/secret/log.yaml'. For configure log level of
            console output used 'log_level' args

        Args:
            log_level: log level for console output
            file_path: path of custom log config

        """
        try:
            level = Logger.LEVELS.get(log_level.upper())
            if level is None:
                raise ValueError(f"log level given: {log_level} -- must be one of: {' | '.join(Logger.LEVELS.keys())}")
            logging_config = Util.yaml_load(file_path) if file_path else None
            if not logging_config:
                logging_config = Util.yaml_load(APP_PATH / "secret" / "log.yaml")
            log_dir = Path(logging_config["handlers"]["logfile"]["filename"]).resolve().parent
            log_dir.mkdir(exist_ok=True)
            logging_config["handlers"]["console"]["level"] = level
            logging.config.dictConfig(logging_config)
            for module in logging_config["ignore"]:
                logging.getLogger(module).setLevel(logging.ERROR)
        except OSError:
            logging.basicConfig(level=logging.WARNING)



from typing import List

from credsweeper.credentials import Candidate


def augment_candidates(candidates: List[Candidate], new_candidates: List[Candidate]):
    """
    Augments candidates with new_candidates if value of line data is not present in the candidates

    Args:
        candidates: [IN/OUT] list of candidates to be augmented
        new_candidates: [IN] list with new candidates

    """

    if not new_candidates:
        return
    found_values = set(line_data.value for candidate in candidates  #
                       for line_data in candidate.line_data_list)
    for new_candidate in new_candidates:
        for line_data in new_candidate.line_data_list:
            if line_data.value not in found_values:
                candidates.append(new_candidate)
                break



from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueLastWordCheck(Filter):
    """Check that secret is not short value that ends with `:`."""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if 16 > len(line_data.value) and not line_data.is_well_quoted_value and line_data.value.endswith(':'):
            return True
        return False



from typing import List

from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.rules import Rule
from credsweeper.scanner.scan_type import ScanType


class SinglePattern(ScanType):
    """Check if single line rule present in the line."""

    @classmethod
    def run(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Check if regex pattern defined in a rule is present in a line.

        Args:
            config: config object of user configs
            rule: Rule object to check current line
            target: Analysis target

        Return:
            List of Candidate objects if pattern defined in a rule is present in a line
            and filters defined in rule do not remove current line. Empty list - otherwise

        """

        return cls._get_candidates(config, rule, target)



import io
import logging
from abc import ABC
from typing import List, Optional

from pdfminer.high_level import extract_pages
from pdfminer.layout import LAParams, LTText, LTItem

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider, MIN_DATA_LEN
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class PdfScanner(AbstractScanner, ABC):
    """Implements pdf scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan PDF elements recursively and the whole text on page as strings"""
        # PyPDF2 - https://github.com/py-pdf/pypdf/issues/1328 text in table is merged without spaces
        # pdfminer.six - splits text in table to many lines. Allows to walk through elements
        try:
            candidates = []
            for page in extract_pages(io.BytesIO(data_provider.data), laparams=LAParams()):
                for element in page:
                    if isinstance(element, LTText):
                        element_text = element.get_text().strip()
                        if 0 < depth and element_text:
                            if MIN_DATA_LEN < len(element_text):
                                pdf_content_provider = DataContentProvider(
                                    data=element_text.encode(),
                                    file_path=data_provider.file_path,
                                    file_type=data_provider.file_type,
                                    info=f"{data_provider.info}|PDF:{page.pageid}")
                                new_limit = recursive_limit_size - len(pdf_content_provider.data)
                                element_candidates = self.recursive_scan(pdf_content_provider, depth, new_limit)
                                candidates.extend(element_candidates)
                        else:
                            string_data_provider = StringContentProvider(lines=[element_text],
                                                                         file_path=data_provider.file_path,
                                                                         file_type=data_provider.file_type,
                                                                         info=f"{data_provider.info}|PDF:{page.pageid}")
                            pdf_candidates = self.scanner.scan(string_data_provider)
                            candidates.extend(pdf_candidates)
                    elif isinstance(element, LTItem):
                        pass
                    else:
                        logger.error(f"Unsupported {element}")
            return candidates
        except Exception as pdf_exc:
            logger.error(f"{data_provider.file_path}:{pdf_exc}")
        return None



import ast
import base64
import json
import logging
import math
import os
import re
import string
import struct
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional, Union

import whatthepatch
import yaml
from lxml import etree
from typing_extensions import TypedDict

from credsweeper.common.constants import DiffRowType, AVAILABLE_ENCODINGS, \
    DEFAULT_ENCODING, LATIN_1, CHUNK_SIZE, MAX_LINE_LENGTH, CHUNK_STEP_SIZE

logger = logging.getLogger(__name__)

DiffDict = TypedDict(
    "DiffDict",
    {
        "old": Optional[int],  #
        "new": Optional[int],  #
        "line": Union[str, bytes],  # bytes are possibly since whatthepatch v1.0.4
        "hunk": Any  # not used
    })


@dataclass(frozen=True)
class DiffRowData:
    """Class for keeping data of diff row."""

    line_type: DiffRowType
    line_numb: int
    line: str


class Util:
    """Class that contains different useful methods."""

    @staticmethod
    def get_extension(file_path: str, lower=True) -> str:
        """Return extension of file in lower case by default e.g.: '.txt', '.JPG'"""
        _, extension = os.path.splitext(str(file_path))
        return extension.lower() if lower else extension

    @staticmethod
    def get_regex_combine_or(re_strs: List[str]) -> str:
        """Routine combination for regex 'or'"""
        result = "(?:"

        for elem in re_strs:
            result += elem + "|"

        if result[-1] == "|":
            result = result[:-1]
        result += ")"

        return result

    @staticmethod
    def get_shannon_entropy(data: str, iterator: str) -> float:
        """Borrowed from http://blog.dkbza.org/2007/05/scanning-data-for-entropy-anomalies.html."""
        if not data:
            return 0

        entropy = 0.
        data_len = float(len(data))
        for x in iterator:
            p_x = data.count(x) / data_len
            if p_x > 0:
                entropy += -p_x * math.log(p_x, 2)

        return entropy

    """Precalculated data for speedup"""
    MIN_DATA_ENTROPY: Dict[int, float] = {
        16: 1.66973671780348,
        20: 2.07723544540831,
        32: 3.25392803184602,
        40: 3.64853567064867,
        64: 4.57756933688035,
        384: 7.39,
        512: 7.55,
    }

    @staticmethod
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal entropy for size of random data. Precalculated data is applied for speedup"""
        if x in Util.MIN_DATA_ENTROPY:
            y = Util.MIN_DATA_ENTROPY[x]
        elif 8 < x < 64:
            # approximated for range 12 - 64
            _x = x - 8
            y = ((0.000016617804 * _x - 0.002695077) * _x + 0.170393) * _x + 0.4
        elif 64 < x < 384:
            # logarithm base 2 - slow, but precise
            _x = x - 8
            y = 1.095884 * math.log2(_x) - 1.90156
        elif 384 < x < 512:
            # solved for 384 - 512
            y = -0.11215851 * math.log2(x)**2 + 2.34303484 * math.log2(x) - 4.4466237
        else:
            # less or equal to 8 bytes might have 0 entropy
            y = 0
        return y

    @staticmethod
    def is_ascii_entropy_validate(data: bytes) -> bool:
        """
        Tests small data sequence (<256) for data randomness by testing for ascii and shannon entropy
        Returns True when data is an ASCII symbols or have small entropy
        """
        if not data:
            return True
        data_len = len(data)
        if 9 > data_len:
            # even random data may have 0 entropy for length of 8 bytes and less
            return True
        entropy = 0.
        cells = [int(0)] * 256
        ascii_test = True
        # "basket" sorting approach
        for x in data:
            cells[x] += 1
            if ascii_test and 0b10000000 & x:
                ascii_test = False
        if ascii_test:
            # only ascii symbols found
            return True
        left = 0.
        step = 256.0 / data_len
        right = left + step
        while left < 256:
            cell_sum = 0
            i = int(left)
            r = int(right)
            while i < r and i < 256:
                cell_sum += cells[i]
                i += 1
            p_x = float(cell_sum) / data_len
            if p_x > 0:
                entropy += -p_x * math.log2(p_x)
            left = right
            right += step
        min_entropy = Util.get_min_data_entropy(data_len)
        return entropy < min_entropy

    @staticmethod
    def is_binary(data: bytes) -> bool:
        """
        Returns true if any recognized binary format found
        or two zeroes sequence is found which never exists in text format (UTF-8, UTF-16)
        UTF-32 is not supported
        """
        if Util.is_zip(data) \
                or Util.is_gzip(data) \
                or Util.is_tar(data) \
                or Util.is_bzip2(data) \
                or Util.is_pdf(data) \
                or Util.is_elf(data):
            return True
        if 0 <= data.find(b"\0\0", 0, MAX_LINE_LENGTH):
            return True
        non_ascii_cnt = 0
        for n in range(min([len(data), MAX_LINE_LENGTH])):
            i = data[n]
            if 0x20 > i and i not in (0x09, 0x0A, 0x0D) or 0x7E < i < 0xA0:
                # less than space and not tab, line feed, line end
                non_ascii_cnt += 1
        if data:
            # experiment for 255217 binary files shown avg = 0.268264 ± 0.168767, so let choose minimal
            chunk_len = float(MAX_LINE_LENGTH if MAX_LINE_LENGTH < len(data) else len(data))
            result = 0.1 < non_ascii_cnt / chunk_len
        else:
            # empty data case
            result = False
        return result

    @staticmethod
    def read_file(path: Union[str, Path], encodings: Optional[List[str]] = None) -> List[str]:
        """Read the file content using different encodings.

        Try to read the contents of the file according to the list of encodings "encodings" as soon as reading
        occurs without any exceptions, the data is returned in the current encoding

        Args:
            path: path to file
            encodings: supported encodings

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned

        """
        data = Util.read_data(path)
        return Util.decode_bytes(data, encodings)

    @staticmethod
    def decode_text(content: bytes, encodings: Optional[List[str]] = None) -> Optional[str]:
        """Decode content using different encodings.

        Try to decode bytes according to the list of encodings "encodings"
        occurs without any exceptions. UTF-16 requires BOM

        Args:
            content: raw data that might be text
            encodings: supported encodings

        Return:
            Decoded text in str for any suitable encoding
            or None when binary data detected

        """
        text = None
        binary_suggest = False
        if encodings is None:
            encodings = AVAILABLE_ENCODINGS
        for encoding in encodings:
            try:
                if binary_suggest and LATIN_1 == encoding and Util.is_binary(content):
                    # LATIN_1 may convert data (bytes in range 0x80:0xFF are transformed)
                    # so skip this encoding when checking binaries
                    logger.warning("Binary file detected")
                    break
                text = content.decode(encoding, errors="strict")
                if content != text.encode(encoding, errors="strict"):
                    # the check helps to detect a real encoding
                    raise UnicodeError
                break
            except UnicodeError:
                binary_suggest = True
                logger.info(f"UnicodeError: Can't decode content as {encoding}.")
            except Exception as exc:
                logger.error(f"Unexpected Error: Can't read content as {encoding}. Error message: {exc}")
        return text

    @staticmethod
    def decode_bytes(content: bytes, encodings: Optional[List[str]] = None) -> List[str]:
        """Decode content using different encodings.

        Try to decode bytes according to the list of encodings "encodings"
        occurs without any exceptions. UTF-16 requires BOM

        Args:
            content: raw data that might be text
            encodings: supported encodings

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned
            Also empty list will be returned after last encoding and 0 symbol is present in lines not at end

        """
        if text := Util.decode_text(content, encodings):
            lines = text.replace('\r\n', '\n').replace('\r', '\n').split('\n')
        else:
            lines = []
        return lines

    @staticmethod
    def patch2files_diff(raw_patch: List[str], change_type: DiffRowType) -> Dict[str, List[DiffDict]]:
        """Generate files changes from patch for added or deleted filepaths.

        Args:
            raw_patch: git patch file content
            change_type: change type to select, DiffRowType.ADDED or DiffRowType.DELETED

        Return:
            return dict with ``{file paths: list of file row changes}``, where
            elements of list of file row changes represented as::

                {
                    "old": line number before diff,
                    "new": line number after diff,
                    "line": line text,
                    "hunk": diff hunk number
                }

        """
        if not raw_patch:
            return {}

        added_files, deleted_files = {}, {}
        try:
            for patch in whatthepatch.parse_patch(raw_patch):
                if patch.changes is None:
                    logger.warning(f"Patch '{str(patch.header)}' cannot be scanned")
                    continue
                changes = []
                for change in patch.changes:
                    change_dict = change._asdict()
                    changes.append(change_dict)

                added_files[patch.header.new_path] = changes
                deleted_files[patch.header.old_path] = changes
            if change_type == DiffRowType.ADDED:
                return added_files
            elif change_type == DiffRowType.DELETED:
                return deleted_files
            else:
                logger.error(f"Change type should be one of: '{DiffRowType.ADDED}', '{DiffRowType.DELETED}';"
                             f" but received {change_type}")
        except Exception as exc:
            logger.exception(exc)
        return {}

    @staticmethod
    def preprocess_diff_rows(
            added_line_number: Optional[int],  #
            deleted_line_number: Optional[int],  #
            line: str) -> List[DiffRowData]:
        """Auxiliary function to extend diff changes.

        Args:
            added_line_number: number of added line or None
            deleted_line_number: number of deleted line or None
            line: the text line

        Return:
            diff rows data with as list of row change type, line number, row content

        """
        rows_data: List[DiffRowData] = []
        if isinstance(added_line_number, int):
            # indicates line was inserted
            rows_data.append(DiffRowData(DiffRowType.ADDED, added_line_number, line))
        if isinstance(deleted_line_number, int):
            # indicates line was removed
            rows_data.append(DiffRowData(DiffRowType.DELETED, deleted_line_number, line))
        return rows_data

    @staticmethod
    def wrong_change(change: DiffDict) -> bool:
        """Returns True if the change is wrong"""
        for i in ["line", "new", "old"]:
            if i not in change:
                logger.error(f"Skipping wrong change {change}")
                return True
        return False

    @staticmethod
    def preprocess_file_diff(changes: List[DiffDict]) -> List[DiffRowData]:
        """Generate changed file rows from diff data with changed lines (e.g. marked + or - in diff).

        Args:
            changes: git diff by file rows data

        Return:
            diff rows data with as list of row change type, line number, row content

        """
        if not changes:
            return []

        rows_data = []
        # process diff to restore lines and their positions
        for change in changes:
            if Util.wrong_change(change):
                continue
            line = change["line"]
            if isinstance(line, str):
                rows_data.extend(Util.preprocess_diff_rows(change.get("new"), change.get("old"), line))
            elif isinstance(line, bytes):
                logger.warning("The feature is available with the deep scan option")
            else:
                logger.error(f"Unknown type of line {type(line)}")

        return rows_data

    @staticmethod
    def is_zip(data: bytes) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures"""
        if isinstance(data, bytes) and 3 < len(data):
            # PK
            if data.startswith(b"PK"):
                if 0x03 == data[2] and 0x04 == data[3]:
                    return True
                # empty archive - no sense to scan
                elif 0x05 == data[2] and 0x06 == data[3]:
                    return True
                # spanned archive - NOT SUPPORTED
                elif 0x07 == data[2] and 0x08 == data[3]:
                    return False
        return False

    @staticmethod
    def is_tar(data: bytes) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures"""
        if isinstance(data, bytes) and 512 <= len(data):
            if 0x75 == data[257] and 0x73 == data[258] and 0x74 == data[259] \
                    and 0x61 == data[260] and 0x72 == data[261] and (
                    0x00 == data[262] and 0x30 == data[263] and 0x30 == data[264]
                    or
                    0x20 == data[262] and 0x20 == data[263] and 0x00 == data[264]
            ):
                try:
                    chksum = tarfile.nti(data[148:156])  # type: ignore
                    unsigned_chksum, signed_chksum = tarfile.calc_chksums(data)  # type: ignore
                    return bool(chksum == unsigned_chksum or chksum == signed_chksum)
                except Exception as exc:
                    logger.exception(f"Corrupted TAR ? {exc}")
        return False

    @staticmethod
    def is_bzip2(data: bytes) -> bool:
        """According https://en.wikipedia.org/wiki/Bzip2"""
        if isinstance(data, bytes) and 10 <= len(data):
            if data.startswith(b"\x42\x5A\x68") \
                    and 0x31 <= data[3] <= 0x39 \
                    and 0x31 == data[4] and 0x41 == data[5] and 0x59 == data[6] \
                    and 0x26 == data[7] and 0x53 == data[8] and 0x59 == data[9]:
                return True
        return False

    @staticmethod
    def is_gzip(data: bytes) -> bool:
        """According https://www.rfc-editor.org/rfc/rfc1952"""
        if isinstance(data, bytes) and 3 <= len(data):
            if data.startswith(b"\x1F\x8B\x08"):
                return True
        return False

    @staticmethod
    def is_pdf(data: bytes) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - pdf"""
        if isinstance(data, bytes) and 5 <= len(data):
            if data.startswith(b"\x25\x50\x44\x46\x2D"):
                return True
        return False

    @staticmethod
    def is_jks(data: bytes) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - jks"""
        if isinstance(data, bytes) and 4 <= len(data):
            if data.startswith(b"\xFE\xED\xFE\xED"):
                return True
        return False

    @staticmethod
    def is_asn1(data: bytes) -> bool:
        """Only sequence type 0x30 and size correctness is checked"""
        data_length = len(data)
        if isinstance(data, bytes) and 4 <= data_length:
            # sequence
            if 0x30 == data[0]:
                # https://www.oss.com/asn1/resources/asn1-made-simple/asn1-quick-reference/basic-encoding-rules.html#Lengths
                length = data[1]
                byte_len = (0x7F & length)
                if 0x80 == length and data.endswith(b"\x00\x00"):
                    return True
                elif 0x80 < length and 1 < byte_len < data_length:  # additional check
                    len_bytes = data[2:2 + byte_len]
                    try:
                        long_size = struct.unpack(">h", len_bytes)
                    except struct.error:
                        long_size = (-1,)  # yapf: disable
                    length = long_size[0]
                elif 0x80 < length and 1 == byte_len:  # small size
                    length = data[2]
                else:
                    byte_len = 0
                return data_length == length + 2 + byte_len
        return False

    @staticmethod
    def is_elf(data: Union[bytes, bytearray]) -> bool:
        """According to https://en.wikipedia.org/wiki/Executable_and_Linkable_Format use only 5 bytes"""
        if isinstance(data, (bytes, bytearray)) and 127 <= len(data):
            # minimal is 127 bytes https://github.com/tchajed/minimal-elf
            if data.startswith(b"\x7f\x45\x4c\x46") and (0x01 == data[5] or 0x02 == data[5]):
                return True
        return False

    @staticmethod
    def is_html(data: Union[bytes, bytearray]) -> bool:
        """Used to detect html format. Suppose, invocation of is_xml() was True before."""
        if isinstance(data, (bytes, bytearray)):
            for opening_tag, closing_tag in [(b"<html>", b"</html>"), (b"<table", b"</table>"), (b"<p>", b"</p>"),
                                             (b"<span>", b"</span>"), (b"<div>", b"</div>"), (b"<li>", b"</li>"),
                                             (b"<ol>", b"</ol>"), (b"<ul>", b"</ul>"), (b"<th>", b"</th>"),
                                             (b"<tr>", b"</tr>"), (b"<td>", b"</td>")]:
                opening_pos = data.find(opening_tag, 0, MAX_LINE_LENGTH)
                if 0 <= opening_pos < data.find(closing_tag, opening_pos):
                    # opening and closing tags were found - suppose it is an HTML
                    return True
        return False

    @staticmethod
    def is_mxfile(data: Union[bytes, bytearray]) -> bool:
        """Used to detect mxfile (drawio) format. Suppose, invocation of is_xml() was True before."""
        if isinstance(data, (bytes, bytearray)):
            mxfile_tag_pos = data.find(b"<mxfile", 0, MAX_LINE_LENGTH)
            if 0 <= mxfile_tag_pos < data.find(b"</mxfile>", mxfile_tag_pos):
                return True
        return False

    # A well-formed XML must start from < or a whitespace character
    XML_FIRST_BRACKET_PATTERN = re.compile(rb"^\s*<")
    XML_OPENING_TAG_PATTERN = re.compile(rb"<([0-9A-Za-z_]{1,256})")

    @staticmethod
    def is_xml(data: Union[bytes, bytearray]) -> bool:
        """Used to detect xml format from raw bytes"""
        if isinstance(data, (bytes, bytearray)) and Util.XML_FIRST_BRACKET_PATTERN.search(data, 0, MAX_LINE_LENGTH):
            if first_bracket_match := Util.XML_OPENING_TAG_PATTERN.search(data, 0, MAX_LINE_LENGTH):
                start_pos = first_bracket_match.start()
                closing_tag = b"</" + first_bracket_match.group(1) + b">"
                if start_pos < data.find(closing_tag, start_pos):
                    return True
        return False

    @staticmethod
    def is_eml(data: Union[bytes, bytearray]) -> bool:
        """According to https://datatracker.ietf.org/doc/html/rfc822 lookup the fields: Date, From, To or Subject"""
        if isinstance(data, (bytes, bytearray)):
            if ((b"\nDate:" in data or data.startswith(b"Date:"))  #
                    and (b"\nFrom:" in data or data.startswith(b"From:"))  #
                    and (b"\nTo:" in data or data.startswith(b"To:")  #
                         or b"\nSubject:" in data or data.startswith(b"Subject:"))):
                return True
        return False

    @staticmethod
    def read_data(path: Union[str, Path]) -> Optional[bytes]:
        """Read the file bytes as is.

        Try to read the data of the file.

        Args:
            path: path to file

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned

        """

        try:
            with open(path, "rb") as file:
                return file.read()
        except Exception as exc:
            logger.error(f"Unexpected Error: Can not read '{path}'. Error message: '{exc}'")
        return None

    @staticmethod
    def get_xml_from_lines(xml_lines: List[str]) -> Tuple[Optional[List[str]], Optional[List[int]]]:
        """Parse xml data from list of string and return List of str.

        Args:
            xml_lines: list of lines of xml data

        Return:
            List of formatted string(f"{root.tag} : {root.text}")

        Raises:
            xml exception

        """
        lines = []
        line_nums = []
        tree = etree.fromstringlist(xml_lines)
        for element in tree.iter():
            tag = Util._extract_element_data(element, "tag")
            text = Util._extract_element_data(element, "text")
            lines.append(f"{tag} : {text}")
            line_nums.append(element.sourceline)
        return lines, line_nums

    @staticmethod
    def _extract_element_data(element, attr) -> str:
        """Extract xml element data to string.

        Try to extract the xml data and strip() the string.

        Args:
            element: xml element
            attr: attribute name

        Return:
            String xml data with strip()

        """
        element_attr: Any = getattr(element, attr)
        if element_attr is None or not isinstance(element_attr, str):
            return ""
        return str(element_attr).strip()

    @staticmethod
    def json_load(file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> Any:
        """Load dictionary from json file"""
        try:
            with open(file_path, "r", encoding=encoding) as f:
                return json.load(f)
        except Exception as exc:
            logging.error(f"Failed to read: {file_path} {exc}")
        return None

    @staticmethod
    def json_dump(obj: Any, file_path: Union[str, Path], encoding=DEFAULT_ENCODING, indent=4) -> None:
        """Write dictionary to json file"""
        try:
            with open(file_path, "w", encoding=encoding) as f:
                json.dump(obj, f, indent=indent)
        except Exception as exc:
            logging.error(f"Failed to write: {file_path} {exc}")

    @staticmethod
    def yaml_load(file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> Any:
        """Load dictionary from yaml file"""
        try:
            with open(file_path, "r", encoding=encoding) as f:
                return yaml.load(f, Loader=yaml.FullLoader)
        except Exception as exc:
            logger.error(f"Failed to read {file_path} {exc}")
        return None

    @staticmethod
    def yaml_dump(obj: Any, file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> None:
        """Write dictionary to yaml file"""
        try:
            with open(file_path, "w", encoding=encoding) as f:
                yaml.dump(obj, f)
        except Exception as exc:
            logging.error(f"Failed to write: {file_path} {exc}")

    @staticmethod
    def __extract_value(node: Any, value: Any) -> List[Any]:
        result = []
        for i in getattr(node, "targets"):
            if hasattr(i, "id"):
                result.append({getattr(i, "id"): value})
            else:
                logger.error(f"{str(i)} has no 'id'")
        return result

    @staticmethod
    def __extract_assign(node: Any) -> List[Any]:
        result = []
        if hasattr(node, "value") and hasattr(node, "targets"):
            value = getattr(node, "value")
            if hasattr(value, "value"):
                # python 3.8 - 3.10
                result.extend(Util.__extract_value(node, getattr(value, "value")))
            else:
                logger.error(f"value.{value} has no 'value' {dir(value)}")
        else:
            logger.error(f"{str(node)} has no 'value' {dir(node)}")
        return result

    @staticmethod
    def ast_to_dict(node: Any) -> List[Any]:
        """Recursive parsing AST tree of python source to list with strings"""
        result: List[Any] = []
        if hasattr(node, "value") and isinstance(node.value, str):
            result.append(node.value)

        if isinstance(node, ast.Module) \
                or isinstance(node, ast.FunctionDef):
            if hasattr(node, "body"):
                for i in node.body:
                    x = Util.ast_to_dict(i)
                    if x:
                        result.extend(x)
        elif isinstance(node, ast.Import):
            logger.debug("Import:%s", str(node))
        elif isinstance(node, ast.Assign):
            result.extend(Util.__extract_assign(node))
        elif isinstance(node, ast.Expr) \
                or isinstance(node, ast.AnnAssign) \
                or isinstance(node, ast.AugAssign) \
                or isinstance(node, ast.Call) \
                or isinstance(node, ast.JoinedStr) \
                or isinstance(node, ast.Return) \
                or isinstance(node, ast.ImportFrom) \
                or isinstance(node, ast.Assert) \
                or isinstance(node, ast.Pass) \
                or isinstance(node, ast.Raise) \
                or isinstance(node, ast.Str) \
                or isinstance(node, ast.Name) \
                or isinstance(node, ast.FormattedValue) \
                or isinstance(node, ast.Global):
            if hasattr(node, "value"):
                result.extend(Util.ast_to_dict(getattr(node, "value")))
            if hasattr(node, "args"):
                for i in getattr(node, "args"):
                    result.extend(Util.ast_to_dict(i))
            if hasattr(node, "values"):
                for i in getattr(node, "values"):
                    result.extend(Util.ast_to_dict(i))
            else:
                logger.debug(f"skip:{str(node)}")
        else:
            logger.debug(f"unknown:{str(node)}")
        return result

    @staticmethod
    def parse_python(source: str) -> List[Any]:
        """Parse python source to list of strings and assignments"""
        src = ast.parse(source)
        result = Util.ast_to_dict(src)
        return result

    @staticmethod
    def decode_base64(text: str, padding_safe: bool = False, urlsafe_detect=False) -> bytes:
        """decode text to bytes with / without padding detect and urlsafe symbols"""
        value = text
        if padding_safe:
            pad_num = 0x3 & len(value)
            if pad_num:
                value += '=' * (4 - pad_num)
        if urlsafe_detect and ('-' in value or '_' in value):
            decoded = base64.b64decode(value, altchars=b"-_", validate=True)
        else:
            decoded = base64.b64decode(value, validate=True)
        return decoded

    @staticmethod
    def get_chunks(line_len: int) -> List[Tuple[int, int]]:
        """Returns chunks positions for given line length"""
        # line length is over MAX_LINE_LENGTH already
        chunks = [(0, CHUNK_SIZE)]
        # case for oversize line
        next_offset = CHUNK_STEP_SIZE
        while line_len > next_offset:
            # the target is too long for single "finditer" - it will be scanned by chunks
            if line_len > next_offset + MAX_LINE_LENGTH:
                # the chunk is not the before last
                chunks.append((next_offset, next_offset + CHUNK_SIZE))
                next_offset += CHUNK_STEP_SIZE
            else:
                # the tail of line is between CHUNK_SIZE and MAX_LINE_LENGTH
                chunks.append((next_offset, line_len))
                break
        return chunks

    @staticmethod
    def subtext(text: str, pos: int, hunk_size: int) -> str:
        """cut text symmetrically for given position or use remained quota to be fitted in 2x hunk_size"""
        if hunk_size <= pos:
            left_quota = 0
            left_pos = pos - hunk_size
        else:
            left_quota = hunk_size - pos
            left_pos = 0
        # skip leading whitespaces in result string
        for i in range(left_pos, pos):
            if text[i] in string.whitespace:
                left_quota += 1
                left_pos += 1
            else:
                break
        right_remain = len(text) - pos
        if hunk_size <= right_remain:
            right_quota = 0
            right_pos = pos + hunk_size + left_quota
        else:
            right_quota = hunk_size - right_remain
            right_pos = pos + hunk_size + left_quota
        if len(text) < right_pos:
            right_pos = len(text)
        if 0 < left_pos:
            left_pos -= right_quota
            if 0 > left_pos:
                left_pos = 0
        return text[left_pos:right_pos].rstrip()

    @staticmethod
    def get_excel_column_name(column_index: int) -> str:
        """Converts index based column position into Excel style column name"""
        name = ''
        if isinstance(column_index, int):
            while 0 <= column_index:
                column_index, remain = divmod(column_index, 26)
                name = f"{chr(ord('A') + remain)}{name}"
                column_index -= 1
        return name



from abc import abstractmethod
from typing import List, Any, Tuple, Set

import numpy as np

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature


class WordIn(Feature):
    """Abstract feature returns array with all matched words in a string"""

    def __init__(self, words: List[str]):
        super().__init__()
        self.dimension = len(words)
        self.words = sorted(list(set(words)))
        self.enumerated_words = list(enumerate(self.words))
        if len(self.enumerated_words) != self.dimension:
            raise RuntimeError(f"Check duplicates:{words}")

    @property
    def enumerated_words(self) -> List[Tuple[int, str]]:
        """getter for speedup"""
        return self.__enumerated_words

    @enumerated_words.setter
    def enumerated_words(self, enumerated_words: List[Tuple[int, str]]) -> None:
        """setter for speedup"""
        self.__enumerated_words = enumerated_words

    @property
    def dimension(self) -> int:
        """getter"""
        return self.__dimension

    @dimension.setter
    def dimension(self, dimension: int) -> None:
        """setter"""
        self.__dimension = dimension

    @abstractmethod
    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError

    def word_in_str(self, a_string: str) -> np.ndarray:
        """Returns array with words included in a string"""
        result: np.ndarray = np.zeros(shape=[self.dimension], dtype=np.int8)
        for i, word in self.enumerated_words:
            if word in a_string:
                result[i] = 1
        return np.array([result])

    def word_in_set(self, a_strings_set: Set[str]) -> np.ndarray:
        """Returns array with words matches in a_strings_set"""
        result: np.ndarray = np.zeros(shape=[self.dimension], dtype=np.int8)
        for i, word in self.enumerated_words:
            if word in a_strings_set:
                result[i] = 1
        return np.array([result])



import binascii
import contextlib
import struct

from credsweeper.common.constants import ASCII
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueGrafanaServiceCheck(Filter):
    """Check that candidate have a known structure"""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        with contextlib.suppress(Exception):
            checksum = struct.unpack("<I", bytes.fromhex(line_data.value[38:]))[0]
            data = line_data.value[:37].encode(ASCII)
            crc32 = binascii.crc32(data)
            if checksum == crc32:
                return False
        return True



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueArrayDictionaryCheck(Filter):
    """Match call to dictionary or array element.

    This filter checks only calls, not declarations:
        `token = values[i]` would be filtered
        `token = {'root'}` would be kept
    """

    PATTERN = re.compile(r"\[('|\")?.+('|\")?\]")

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.is_well_quoted_value:
            return False
        if self.PATTERN.search(line_data.value):
            return True
        if line_data.wrap and not line_data.is_well_quoted_value and ('[' in line_data.wrap or '(' in line_data.wrap):
            return True

        return False



from credsweeper.filters.filter import Filter  # isort:skip

from credsweeper.filters.line_git_binary_check import LineGitBinaryCheck
from credsweeper.filters.line_specific_key_check import LineSpecificKeyCheck
from credsweeper.filters.line_uue_part_check import LineUUEPartCheck
from credsweeper.filters.value_allowlist_check import ValueAllowlistCheck
from credsweeper.filters.value_array_dictionary_check import ValueArrayDictionaryCheck
from credsweeper.filters.value_atlassian_token_check import ValueAtlassianTokenCheck
from credsweeper.filters.value_azure_token_check import ValueAzureTokenCheck
from credsweeper.filters.value_base32_data_check import ValueBase32DataCheck
from credsweeper.filters.value_base64_data_check import ValueBase64DataCheck
from credsweeper.filters.value_base64_encoded_pem_check import ValueBase64EncodedPem
from credsweeper.filters.value_base64_key_check import ValueBase64KeyCheck
from credsweeper.filters.value_base64_part_check import ValueBase64PartCheck
from credsweeper.filters.value_blocklist_check import ValueBlocklistCheck
from credsweeper.filters.value_camel_case_check import ValueCamelCaseCheck
from credsweeper.filters.value_couple_keyword_check import ValueCoupleKeywordCheck
from credsweeper.filters.value_dictionary_keyword_check import ValueDictionaryKeywordCheck
from credsweeper.filters.value_dictionary_value_length_check import ValueDictionaryValueLengthCheck
from credsweeper.filters.value_discord_bot_check import ValueDiscordBotCheck
from credsweeper.filters.value_entropy_base32_check import ValueEntropyBase32Check
from credsweeper.filters.value_entropy_base36_check import ValueEntropyBase36Check
from credsweeper.filters.value_entropy_base64_check import ValueEntropyBase64Check
from credsweeper.filters.value_file_path_check import ValueFilePathCheck
from credsweeper.filters.value_first_word_check import ValueFirstWordCheck
from credsweeper.filters.value_github_check import ValueGitHubCheck
from credsweeper.filters.value_grafana_check import ValueGrafanaCheck
from credsweeper.filters.value_grafana_service_check import ValueGrafanaServiceCheck
from credsweeper.filters.value_hex_number_check import ValueHexNumberCheck
from credsweeper.filters.value_jfrog_token_check import ValueJfrogTokenCheck
from credsweeper.filters.value_json_web_token_check import ValueJsonWebTokenCheck
from credsweeper.filters.value_last_word_check import ValueLastWordCheck
from credsweeper.filters.value_method_check import ValueMethodCheck
from credsweeper.filters.value_not_allowed_pattern_check import ValueNotAllowedPatternCheck
from credsweeper.filters.value_not_part_encoded_check import ValueNotPartEncodedCheck
from credsweeper.filters.value_number_check import ValueNumberCheck
from credsweeper.filters.value_pattern_check import ValuePatternCheck
from credsweeper.filters.value_similarity_check import ValueSimilarityCheck
from credsweeper.filters.value_split_keyword_check import ValueSplitKeywordCheck
from credsweeper.filters.value_string_type_check import ValueStringTypeCheck
from credsweeper.filters.value_token_base32_check import ValueTokenBase32Check
from credsweeper.filters.value_token_base36_check import ValueTokenBase36Check
from credsweeper.filters.value_token_base64_check import ValueTokenBase64Check
from credsweeper.filters.value_token_check import ValueTokenCheck



from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueStringTypeCheck(Filter):
    r"""Check if line_data is in source code file that require quotes for string declaration.

    If it is, then checks if line_data really have string literal declaration.
    Comment rows in source files (start with //, /\*, etc) ignored.

    True if:

    - line_data have no value
    - line_data have no path
    - line_data is in source code file (.cpp, .py, etc.) and is not comment
      and contain no quotes (so no string literal declared)

    False otherwise
    """

    def __init__(self, config: Config) -> None:
        self.check_for_literals = config.check_for_literals

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if not self.check_for_literals or line_data.url_part:
            return False

        not_quoted = not line_data.is_well_quoted_value
        not_comment = not line_data.is_comment()

        if line_data.is_source_file_with_quotes() and not_comment and not_quoted and not line_data.is_quoted \
                and line_data.separator and '=' in line_data.separator:
            # heterogeneous code e.g. YAML in Python uses colon sign instead equals
            return True

        return False



from typing import Union

from credsweeper.common import static_keyword_checklist
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueSplitKeywordCheck(Filter):
    """Check value by splitting with standard whitespace separators and any word is not matched in checklist."""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        words: Union[set, list] = line_data.value.lower().split()
        keyword_set = static_keyword_checklist.keyword_set
        for word in words:
            if word in keyword_set:
                return True
        return False



from abc import ABC, abstractmethod
from typing import List, Any

import numpy as np

from credsweeper.credentials import Candidate


class Feature(ABC):
    """Base class for features."""

    def __init__(self):
        self.words = []

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        """Call base class for features.

        Args:
            candidates: list of candidates to extract features

        """
        return np.array([self.extract(candidate) for candidate in candidates])

    @abstractmethod
    def extract(self, candidate: Candidate) -> Any:
        """Abstract method of base class"""
        raise NotImplementedError

    @property
    def words(self) -> List[str]:
        """getter"""
        return self.__words

    @words.setter
    def words(self, words: List[str]) -> None:
        """setter"""
        self.__words = words

    def any_word_in_(self, a_string: str) -> bool:
        """Returns true if any words in a string"""
        for i in self.words:
            if i in a_string:
                return True
        return False



import logging
from functools import cached_property
from typing import Optional, Any, Generator

from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider

logger = logging.getLogger(__name__)


class StructContentProvider(ContentProvider):
    """Content provider to keep structured data"""

    def __init__(
            self,  #
            struct: Any,  #
            file_path: Optional[str] = None,  #
            file_type: Optional[str] = None,  #
            info: Optional[str] = None) -> None:
        """
        Parameters:
            struct: Various structure (string, dictionary, list)

        """
        super().__init__(file_path=file_path, file_type=file_type, info=info)
        self.__struct = struct

    @cached_property
    def data(self) -> bytes:
        """data getter for StructContentProvider"""
        raise NotImplementedError(__name__)

    @cached_property
    def struct(self) -> Any:
        """struct getter for StructContentProvider"""
        return self.__struct

    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        self.__struct = None
        if hasattr(self, "struct"):
            delattr(self, "struct")

    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Return nothing. The class provides only data storage.

        Args:
            min_len: minimal line length to scan

        Raise:
            NotImplementedError

        """
        raise NotImplementedError()



import io
import logging
from abc import ABC
from typing import List, Optional

from pptx import Presentation

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class PptxScanner(AbstractScanner, ABC):
    """Implements pptx scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan pptx text elements for all slides"""
        try:
            candidates = []
            pptx_lines = []
            presentation = Presentation(io.BytesIO(data_provider.data))
            for n, slide in enumerate(presentation.slides):
                for shape in slide.shapes:
                    if shape.has_text_frame:
                        for paragraph in shape.text_frame.paragraphs:
                            pptx_lines.append(paragraph.text)
                string_data_provider = StringContentProvider(lines=pptx_lines,
                                                             file_path=data_provider.file_path,
                                                             file_type=data_provider.file_type,
                                                             info=f"{data_provider.info}|PPTX:{n+1}")
                pptx_candidates = self.scanner.scan(string_data_provider)
                candidates.extend(pptx_candidates)
            return candidates
        except Exception as pptx_exc:
            logger.error(f"{data_provider.file_path}:{pptx_exc}")
        return None



from functools import cached_property
from typing import Set, List

from credsweeper.app import APP_PATH


class KeywordChecklist:
    """KeywordsChecklist contains words 3 or more letters length"""
    __keyword_set: Set[str]
    __morpheme_set: Set[str]
    KEYWORD_PATH = APP_PATH / "common" / "keyword_checklist.txt"
    MORPHEME_PATH = APP_PATH / "common" / "morpheme_checklist.txt"

    def __init__(self) -> None:
        # used suggested text read style. split() is preferred because it strips 0x0A on end the file
        self.__keyword_list = self.KEYWORD_PATH.read_text().split()
        self.__keyword_list.sort(key=str.__len__, reverse=True)
        self.__keyword_set = set(self.KEYWORD_PATH.read_text().split())
        # The list of morphemes can be combined to form words.
        # The value is considered a variable if at least two exist.
        self.__morpheme_set = set(self.MORPHEME_PATH.read_text().split())

    @cached_property
    def keyword_set(self) -> Set[str]:
        """Get set with keywords"""
        return self.__keyword_set

    @cached_property
    def keyword_list(self) -> List[str]:
        """Get list with keywords in descended order of length"""
        return self.__keyword_list

    @cached_property
    def keyword_len(self) -> int:
        """Length of keyword_set"""
        return len(self.__keyword_set)

    @cached_property
    def morpheme_set(self) -> Set[str]:
        """Get extended set with keywords.

        Return:
            Extended set of strings

        """
        return self.__morpheme_set

    @cached_property
    def morpheme_len(self) -> int:
        """Length of morpheme_set"""
        return len(self.__morpheme_set)

    def check_morphemes(self, line_lower: str, threshold: int) -> bool:
        """Checks limit of morphemes limit in line.

        Args:
            line_lower: input line - MUST be in lower
            threshold: number of minimal morphemes

        Return:
            True - if number of morphemes exceeds the threshold
        """
        matches = 0
        for keyword in self.morpheme_set:
            if keyword in line_lower:
                matches += 1
                if threshold < matches:
                    return True
        return False



import re


class KeywordPattern:
    """Pattern set of keyword types"""
    key_left = r"(\\[nrt]|%[0-9a-f]{2})?"\
               r"(?P<variable>(([`'\"]+[^:='\"`}<>\\/&?]*|[^:='\"`}<>\s()\\/&?;,%]*)" \
               r"(?P<keyword>"
    # there will be inserted a keyword
    key_right = r")" \
                r"(&(quot|apos);|[^%:='\"`<>{?!&]*)[`'\"]*))"  # <variable>
    separator = r"(\s|\\+[tnr])*\]?(\s|\\+[tnr])*" \
                r"(?P<separator>:( [a-z]{3,9}[?]? )?=|:|=(>|&gt;|\\u0026gt;)|!=|===|==|=|%3d)" \
                r"(\s|\\+[tnr])*"
    # might be curly, square or parenthesis with words before
    wrap = r"(?P<wrap>(" \
           r"(new(\s|\\+[tnr])+)?" \
           r"([0-9a-z_.]|-(>|(&|\\\\*u0026)gt;))*" \
           r"[\[\(\{]"\
           r"(\s|\\+[tnr])*" \
           r"([0-9a-z_]{1,32}=)?" \
           r")+)?"
    string_prefix = r"(((b|r|br|rb|u|f|rf|fr|l|@)(?=(\\*[`'\"])))?"
    left_quote = r"(?P<value_leftquote>((?P<esq>\\{1,8})?([`'\"]|&(quot|apos);)){1,4}))?"
    # Authentication scheme ( oauth | basic | bearer | apikey ) precedes to credential
    auth_keywords = r"( ?(oauth|bot|basic|bearer|apikey|accesskey) )?"
    value = r"(?P<value>" \
            r"(?(value_leftquote)" \
            r"(" \
            r"(?!(?P=value_leftquote))" \
            r"(?(esq)((?!(?P=esq)([`'\"]|&(quot|apos);)).)|((?!(?P=value_leftquote)).)))" \
            r"|" \
            r"(?!&(quot|apos);)" \
            r"(\\+([ tnr]|[^\s`'\"])|[^\s`'\",;\\])" \
            r"){3,8000}" \
            r"|(\{[^}]{3,8000}\})" \
            r"|(<[^>]{3,8000}>)" \
            r")"
    right_quote = r"(?(value_leftquote)" \
                  r"(?P<value_rightquote>(?<!\\)(?P=value_leftquote)|\\$|(?<=[0-9a-z+_/-])$)" \
                  r"|" \
                  r"(?(wrap)[\]\)\},;]))"

    @classmethod
    def get_keyword_pattern(cls, keyword: str) -> re.Pattern:
        """Returns compiled regex pattern"""
        expression = "".join([  #
            cls.key_left,  #
            keyword,  #
            cls.key_right,  #
            cls.separator,  #
            cls.wrap,  #
            cls.string_prefix,  #
            cls.left_quote,  #
            cls.auth_keywords,  #
            cls.value,  #
            cls.right_quote,  #
        ])
        return re.compile(expression, flags=re.IGNORECASE | re.DOTALL)



import re

from credsweeper.common.constants import ML_HUNK
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class LineSpecificKeyCheck(Filter):
    """Check that values from list below is not in candidate line."""

    NOT_ALLOWED = [r"example", r"\benc[\(\[]", r"\btrue\b", r"\bfalse\b"]
    NOT_ALLOWED_PATTERN = re.compile(Util.get_regex_combine_or(NOT_ALLOWED), re.IGNORECASE)

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.line is None:
            return True
        if 0 <= line_data.variable_start:
            # variable may be defined too
            sub_line_start = 0 if ML_HUNK >= line_data.variable_start else line_data.variable_start - ML_HUNK
        else:
            sub_line_start = 0 if ML_HUNK >= line_data.value_start else line_data.value_start - ML_HUNK

        if self.NOT_ALLOWED_PATTERN.search(line_data.line, sub_line_start, line_data.value_end + ML_HUNK):
            return True

        return False



import logging
from functools import cached_property
from typing import List, Tuple, Generator

from credsweeper.common.constants import DiffRowType
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.utils import DiffRowData, Util, DiffDict

logger = logging.getLogger(__name__)


class DiffContentProvider(ContentProvider):
    """Provide data from a single `.patch` file.

    Parameters:
        file_path: path to file
        change_type: set added or deleted file data to scan
        diff: list of file row changes, with base elements represented as::

            {
                "old": line number before diff,
                "new": line number after diff,
                "line": line text,
                "hunk": diff hunk number
            }

    """

    def __init__(
            self,  #
            file_path: str,  #
            change_type: DiffRowType,  #
            diff: List[DiffDict]) -> None:
        super().__init__(file_path=file_path, info=f"{file_path}:{change_type.value}")
        self.__change_type = change_type
        self.__diff = diff

    @cached_property
    def data(self) -> bytes:
        """data getter for DiffContentProvider"""
        raise NotImplementedError(__name__)

    @cached_property
    def diff(self) -> List[DiffDict]:
        """diff getter for DiffContentProvider"""
        return self.__diff

    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        self.__diff = None
        if hasattr(self, "diff"):
            delattr(self, "diff")

    @staticmethod
    def parse_lines_data(change_type: DiffRowType, lines_data: List[DiffRowData]) -> Tuple[List[int], List[str]]:
        """Parse diff lines data.

        Return list of line numbers with change type "self.change_type" and list of all lines in file
            in original order(replaced all lines not mentioned in diff file with blank line)

        Args:
            change_type: set added or deleted file data to scan
            lines_data: data of all rows mentioned in diff file

        Return:
            tuple of line numbers with change type "self.change_type" and all file lines
            in original order(replaced all lines not mentioned in diff file with blank line)

        """
        change_numbs = []
        all_lines = []
        for line_data in lines_data:
            if line_data.line_type == change_type:
                change_numbs.append(line_data.line_numb)
                all_lines.append(line_data.line)
        return change_numbs, all_lines

    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Preprocess file diff data to scan.

        Args:
            min_len: minimal line length to scan

        Return:
            list of analysis targets of every row of file diff corresponding to change type "self.change_type"

        """
        lines_data = Util.preprocess_file_diff(self.__diff)
        change_numbs, all_lines = self.parse_lines_data(self.__change_type, lines_data)
        return self.lines_to_targets(min_len, all_lines, change_numbs)



import re
from typing import Dict, List, Optional, Set, Any

from humanfriendly import parse_size

from credsweeper.common.constants import Severity, DEFAULT_PATTERN_LEN
from credsweeper.utils import Util


class Config:
    """Class that contain configs that can be changed by user."""

    NOT_ALLOWED_PATH = [
        ".*\\.min\\.js", ".*message.*\\.properties", ".*locale.*\\.properties", ".*makefile.*", ".*package-lock\\.json",
        ".*package\\.json", ".*\\.css", ".*\\.scss"
    ]

    def __init__(self, config: Dict[str, Any]) -> None:
        self.exclude_patterns: List[re.Pattern] = [re.compile(pattern) for pattern in config["exclude"]["pattern"]]
        self.exclude_paths: List[str] = config["exclude"]["path"]
        self.exclude_containers: List[str] = config["exclude"]["containers"]
        self.exclude_documents: List[str] = config["exclude"]["documents"]
        self.exclude_extensions: List[str] = config["exclude"]["extension"]
        self.exclude_lines: Set[str] = set(config["exclude"].get("lines", []))
        self.exclude_values: Set[str] = set(config["exclude"].get("values", []))
        self.source_extensions: List[str] = config["source_ext"]
        self.source_quote_ext: List[str] = config["source_quote_ext"]
        self.find_by_ext_list: List[str] = config["find_by_ext_list"]
        self.bruteforce_list: List[str] = config["bruteforce_list"]
        self.check_for_literals: bool = config["check_for_literals"]
        self.not_allowed_path_pattern = re.compile(f"{Util.get_regex_combine_or(self.NOT_ALLOWED_PATH)}",
                                                   flags=re.IGNORECASE)
        self.use_filters: bool = config["use_filters"]
        self.line_data_output: List[str] = config["line_data_output"]
        self.candidate_output: List[str] = config["candidate_output"]
        self.find_by_ext: bool = config["find_by_ext"]
        self.size_limit: Optional[int] = parse_size(config["size_limit"]) if config["size_limit"] is not None else None
        self.depth: int = int(config["depth"])
        self.doc: bool = config["doc"]
        self.severity: Severity = Severity.get(config.get("severity"))

        self.min_keyword_value_length: int = int(config["min_keyword_value_length"])
        self.min_pattern_value_length: int = int(config["min_pattern_value_length"])

        # Trim exclude patterns from space like characters
        self.exclude_lines = set(line.strip() for line in self.exclude_lines)
        self.exclude_values = set(line.strip() for line in self.exclude_values)

        self.pattern_len = config.get("pattern_len", DEFAULT_PATTERN_LEN)



import base64
import contextlib
import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class LineGitBinaryCheck(Filter):
    """Checks that line is not a part of git binary patch"""
    base85string = re.compile(r"^[A-Za-z][0-9A-Za-z!#$%&()*+;<=>?@^_`{|}~-]{6,65}$")

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if not line_data.line:
            return True
        if 66 < target.line_strip_len:
            return False
        line = target.line_strip
        len_line = len(line)

        # https://github.com/git/git/blob/master/base85.c

        if 6 <= len_line and 0 == ((len_line - 1) % 5) and LineGitBinaryCheck.base85string.match(line):
            size = ord(line[0])
            if 65 <= size <= 90:  # A-Z
                size -= 64
            elif 97 <= size <= 122:  # a-z
                size -= 70
            else:
                return False
            with contextlib.suppress(Exception):
                decoded = base64.b85decode(line[1:])
                return len(decoded) == size

        return False



import binascii
import logging
import os
import sys
import time
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from typing import Any, Union, Dict

from credsweeper import __version__
from credsweeper.app import APP_PATH, CredSweeper
from credsweeper.common.constants import ThresholdPreset, Severity, RuleType, DiffRowType, ML_HUNK
from credsweeper.file_handler.abstract_provider import AbstractProvider
from credsweeper.file_handler.files_provider import FilesProvider
from credsweeper.file_handler.patches_provider import PatchesProvider
from credsweeper.logger.logger import Logger
from credsweeper.utils import Util

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

logger = logging.getLogger(__name__)


def positive_int(value: Any) -> int:
    """Check if number of parallel processes is not a positive number."""
    int_value = int(value)
    if int_value <= 0:
        logger.error("Number of parallel processes should be a positive number: %s", value)
        raise ArgumentTypeError(f"{value} should be greater than 0")
    return int_value


def threshold_or_float(arg: str) -> Union[float, ThresholdPreset]:
    """Return ThresholdPreset or a float from the input string

    Args:
        arg: string that either a float or one of allowed values in ThresholdPreset

    Returns:
        float if arg convertible to float, ThresholdPreset if one of the allowed values

    Raises:
        ArgumentTypeError: if arg cannot be interpreted as float or ThresholdPreset

    """
    allowed_presents = [e.value for e in ThresholdPreset]
    try:
        return float(arg)  # try convert to float
    except ValueError:
        pass
    if arg in allowed_presents:
        return ThresholdPreset[arg]
    raise ArgumentTypeError(f"value must be a float or one of {allowed_presents}")


def logger_levels(log_level: str) -> str:
    """Logger level correctness verification and transformation

    Args:
        log_level: string with level

    Returns True if log_level UPPERCASE is one of keys
    """
    val = log_level.upper()
    if any(val == i for i in Logger.LEVELS.keys()):
        return val
    raise ArgumentTypeError(f"Log level provided: {log_level} -- must be one of: {' | '.join(Logger.LEVELS.keys())}")


def severity_levels(severity_level: str) -> Severity:
    """Severity level correctness verification and transformation

    Args:
        severity_level: string with level

    Returns Severity matched provided string or throws ArgumentTypeError exception
    """

    if severity := Severity.get(severity_level):
        return severity
    raise ArgumentTypeError(
        f"Severity level provided: {severity_level} -- must be one of: {' | '.join([i.value for i in Severity])}")


def check_integrity() -> int:
    """Calculates CRC32 of program files

    Returns CRC32 of files in integer
    """
    crc32 = 0
    for root, dirs, files in os.walk(APP_PATH):
        for file_path in files:
            if Util.get_extension(file_path) in [".py", ".json", ".txt", ".yaml", ".onnx"]:
                data = Util.read_data(os.path.join(root, file_path))
                if data:
                    crc32 ^= binascii.crc32(data)
    return crc32


def get_arguments() -> Namespace:
    """All CLI arguments are defined here"""
    parser = ArgumentParser(prog="python -m credsweeper")
    single_banner_argument = 2 == len(sys.argv) and "--banner" == sys.argv[1]
    group = parser.add_mutually_exclusive_group(required=not single_banner_argument)
    group.add_argument("--path", nargs="+", help="file or directory to scan", dest="path", metavar="PATH")
    group.add_argument("--diff_path", nargs="+", help="git diff file to scan", dest="diff_path", metavar="PATH")
    group.add_argument("--export_config",
                       nargs="?",
                       help="exporting default config to file (default: config.json)",
                       const="config.json",
                       dest="export_config",
                       metavar="PATH")
    group.add_argument("--export_log_config",
                       nargs="?",
                       help="exporting default logger config to file (default: log.yaml)",
                       const="log.yaml",
                       dest="export_log_config",
                       metavar="PATH")
    parser.add_argument("--rules",
                        help="path of rule config file (default: credsweeper/rules/config.yaml). "
                        f"severity:{[i.value for i in Severity]} "
                        f"type:{[i.value for i in RuleType]}",
                        default=None,
                        dest="rule_path",
                        metavar="PATH")
    parser.add_argument("--severity",
                        help=f"set minimum level for rules to apply {[i.value for i in Severity]}"
                        f"(default: '{Severity.INFO}', case insensitive)",
                        default=Severity.INFO,
                        dest="severity",
                        type=severity_levels)
    parser.add_argument("--config",
                        help="use custom config (default: built-in)",
                        default=None,
                        dest="config_path",
                        metavar="PATH")
    parser.add_argument("--log_config",
                        help="use custom log config (default: built-in)",
                        default=None,
                        dest="log_config_path",
                        metavar="PATH")
    parser.add_argument("--denylist",
                        help="path to a plain text file with lines or secrets to ignore",
                        default=None,
                        dest="denylist_path",
                        metavar="PATH")
    parser.add_argument("--find-by-ext",
                        help="find files by predefined extension",
                        dest="find_by_ext",
                        action="store_true")
    parser.add_argument("--depth",
                        help="additional recursive search in data (experimental)",
                        type=positive_int,
                        dest="depth",
                        default=0,
                        required=False,
                        metavar="POSITIVE_INT")
    parser.add_argument("--no-filters", help="disable filters", dest="no_filters", action="store_false")
    parser.add_argument("--doc", help="document-specific scanning", dest="doc", action="store_true")
    parser.add_argument("--ml_threshold",
                        help="setup threshold for the ml model. "
                        "The lower the threshold - the more credentials will be reported. "
                        f"Allowed values: float between 0 and 1, or any of {[e.value for e in ThresholdPreset]} "
                        "(default: medium)",
                        type=threshold_or_float,
                        default=ThresholdPreset.medium,
                        dest="ml_threshold",
                        required=False,
                        metavar="FLOAT_OR_STR")
    parser.add_argument("--ml_batch_size",
                        "-b",
                        help="batch size for model inference (default: 16)",
                        type=positive_int,
                        dest="ml_batch_size",
                        default=16,
                        required=False,
                        metavar="POSITIVE_INT")
    parser.add_argument("--ml_config",
                        help="use external config for ml model",
                        type=str,
                        default=None,
                        dest="ml_config",
                        required=False,
                        metavar="PATH")
    parser.add_argument("--ml_model",
                        help="use external ml model",
                        type=str,
                        default=None,
                        dest="ml_model",
                        required=False,
                        metavar="PATH")
    parser.add_argument("--ml_providers",
                        help="comma separated list of providers for onnx (CPUExecutionProvider is used by default)",
                        type=str,
                        default=None,
                        dest="ml_providers",
                        required=False,
                        metavar="STR")
    parser.add_argument("--jobs",
                        "-j",
                        help="number of parallel processes to use (default: 1)",
                        type=positive_int,
                        dest="jobs",
                        default=1,
                        metavar="POSITIVE_INT")
    parser.add_argument("--thrifty",
                        help="clear objects after scan to reduce memory consumption",
                        action="store_const",
                        const=True)
    parser.add_argument("--skip_ignored",
                        help="parse .gitignore files and skip credentials from ignored objects",
                        dest="skip_ignored",
                        action="store_true")
    parser.add_argument("--save-json",
                        nargs="?",
                        help="save result to json file (default: output.json)",
                        const="output.json",
                        dest="json_filename",
                        metavar="PATH")
    parser.add_argument("--save-xlsx",
                        nargs="?",
                        help="save result to xlsx file (default: output.xlsx)",
                        const="output.xlsx",
                        dest="xlsx_filename",
                        metavar="PATH")
    parser.add_argument("--color", "-C", help="print results with colorization", action="store_const", const=True)
    parser.add_argument("--hashed",
                        help="line, variable, value will be hashed in output",
                        action="store_const",
                        const=True)
    parser.add_argument("--subtext",
                        help=f"line text will be stripped in {2 * ML_HUNK} symbols but value and variable are kept",
                        action="store_const",
                        const=True)
    parser.add_argument("--sort", help="enable output sorting", dest="sort_output", action="store_true")
    parser.add_argument("--log",
                        "-l",
                        help=f"provide logging level of {list(Logger.LEVELS.keys())}"
                        f"(default: 'warning', case insensitive)",
                        default="warning",
                        dest="log",
                        metavar="LOG_LEVEL",
                        type=logger_levels)
    parser.add_argument("--size_limit",
                        help="set size limit of files that for scanning (eg. 1GB / 10MiB / 1000)",
                        dest="size_limit",
                        default=None)
    parser.add_argument("--banner",
                        help="show version and crc32 sum of CredSweeper files at start",
                        action="store_const",
                        const=True)
    parser.add_argument("--version",
                        "-V",
                        help="show program's version number and exit",
                        action="version",
                        version=f"CredSweeper {__version__}")
    return parser.parse_args()


def scan(args: Namespace, content_provider: AbstractProvider) -> int:
    """Scan content_provider data, print results or save them to json_filename is not None

    Args:
        args: arguments of the application
        content_provider: FilesProvider instance to scan data from

    Returns:
        Number of detected credentials

    Warnings:
         DeprecationWarning: Using 'json_filename' and/or 'xlsx_filename' will issue a warning.

    """
    try:
        if args.denylist_path is not None:
            denylist = [line for line in Util.read_file(args.denylist_path) if line]
        else:
            denylist = []

        credsweeper = CredSweeper(rule_path=args.rule_path,
                                  config_path=args.config_path,
                                  json_filename=args.json_filename,
                                  xlsx_filename=args.xlsx_filename,
                                  color=args.color,
                                  hashed=args.hashed,
                                  subtext=args.subtext,
                                  sort_output=args.sort_output,
                                  use_filters=args.no_filters,
                                  pool_count=args.jobs,
                                  ml_batch_size=args.ml_batch_size,
                                  ml_threshold=args.ml_threshold,
                                  ml_config=args.ml_config,
                                  ml_model=args.ml_model,
                                  ml_providers=args.ml_providers,
                                  find_by_ext=args.find_by_ext,
                                  depth=args.depth,
                                  doc=args.doc,
                                  severity=args.severity,
                                  size_limit=args.size_limit,
                                  exclude_lines=denylist,
                                  exclude_values=denylist,
                                  thrifty=args.thrifty,
                                  log_level=args.log)
        return credsweeper.run(content_provider=content_provider)
    except Exception as exc:
        logger.critical(exc, exc_info=True)
    return -1


def main() -> int:
    """Main function"""
    result = EXIT_FAILURE
    start_time = time.time()
    args = get_arguments()
    if args.banner:
        print(f"CredSweeper {__version__} crc32:{check_integrity():08x}")
    Logger.init_logging(args.log, args.log_config_path)
    logger.info(f"Init CredSweeper object with arguments: {args}")
    summary: Dict[str, int] = {}
    if args.path:
        logger.info(f"Run analyzer on path: {args.path}")
        content_provider: AbstractProvider = FilesProvider(args.path, skip_ignored=args.skip_ignored)
        credentials_number = scan(args, content_provider)
        summary["Detected Credentials"] = credentials_number
        if 0 <= credentials_number:
            result = EXIT_SUCCESS
    elif args.diff_path:
        # Analyze added data
        logger.info(f"Run analyzer on added rows from patch files: {args.diff_path}")
        content_provider = PatchesProvider(args.diff_path, change_type=DiffRowType.ADDED)
        add_credentials_number = scan(args, content_provider)
        summary["Added File Credentials"] = add_credentials_number
        # Analyze deleted data
        logger.info(f"Run analyzer on deleted rows from patch files: {args.diff_path}")
        content_provider = PatchesProvider(args.diff_path, change_type=DiffRowType.DELETED)
        del_credentials_number = scan(args, content_provider)
        summary["Deleted File Credentials"] = del_credentials_number
        if 0 <= add_credentials_number and 0 <= del_credentials_number:
            result = EXIT_SUCCESS
    elif args.export_config:
        logging.info(f"Exporting default config to file: {args.export_config}")
        config_dict = Util.json_load(APP_PATH / "secret" / "config.json")
        Util.json_dump(config_dict, args.export_config)
    elif args.export_log_config:
        logging.info(f"Exporting default logger config to file: {args.export_log_config}")
        config_dict = Util.yaml_load(APP_PATH / "secret" / "log.yaml")
        Util.yaml_dump(config_dict, args.export_log_config)
    elif args.banner and 2 == len(sys.argv):
        # only extend version invocation
        result = EXIT_SUCCESS
    else:
        logger.error("Not specified 'path' or 'diff_path'")

    if EXIT_SUCCESS == result and len(summary):
        for k, v in summary.items():
            print(f"{k}: {v}")
        end_time = time.time()
        print(f"Time Elapsed: {end_time - start_time}s")

    return result


if __name__ == "__main__":
    sys.exit(main())



import io
import logging
import os
from pathlib import Path
from typing import List, Dict, Union, Tuple

from git import InvalidGitRepositoryError, NoSuchPathError, Repo

from credsweeper.common.constants import MIN_DATA_LEN
from credsweeper.config import Config
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class FilePathExtractor:
    """Util class to browse files in directories"""

    FIND_BY_EXT_RULE = "Suspicious File Extension"
    located_repos: Dict[Path, Repo] = {}

    @staticmethod
    def apply_gitignore(detected_files: List[str]) -> List[str]:
        """Apply gitignore rules for each file.

        Args:
            detected_files: list of files to be checked

        Return:
            List of files with all files ignored by git removed

        """
        filtered_files = [file_path for file_path in detected_files if FilePathExtractor.is_valid_path(file_path)]

        return filtered_files

    @staticmethod
    def get_file_paths(config: Config, path: Union[str, Path]) -> List[str]:
        """Get all files in the directory. Automatically exclude files non-code or data files (such as .jpg).

        Args:
            config: credsweeper configuration
            path: path to the file or directory to be scanned

        Return:
            List all non-excluded files in the directory

        """
        path = os.path.expanduser(path)  # Replace ~ character with a full path to the home directory
        if not os.path.exists(path):
            logger.warning(f"'{path}' does not exist")
        file_paths = []
        if os.path.isfile(path):
            # suppose, the file is located outside and should be scanned
            if not FilePathExtractor.check_exclude_file(config, path):
                file_paths.append(path)
        elif os.path.isdir(path):
            for dirpath, _, filenames in os.walk(path):
                for filename in filenames:
                    file_path = os.path.join(f"{dirpath}", f"{filename}")
                    if FilePathExtractor.check_exclude_file(config, file_path) or os.path.islink(file_path):
                        continue
                    if os.path.isfile(file_path) and not FilePathExtractor.check_file_size(config, file_path):
                        file_paths.append(file_path)
        else:
            pass  # symbolic links and so on
        return file_paths

    @classmethod
    def is_valid_path(cls, path: str) -> bool:
        """Locate nearest .git directory to the path and check if path is ignored.

        Args:
            path: path to the file or directory to check

        Return:
            False if file is ignored by git. True otherwise

        """
        parent_directory = Path(path).parent

        # Iterate over file path to find nearest ".git" directory
        while True:
            try:
                if parent_directory in cls.located_repos:
                    repo = cls.located_repos[parent_directory]
                else:
                    # The directory must have ".git" in it. If not it occurs error.
                    repo = Repo(parent_directory)

                    # Cache already located repositories, so we would not need to load it for each new file
                    cls.located_repos[parent_directory] = repo

                # Return True if there is no ignored file in 'path' and False if any.
                return len(repo.ignored(path)) == 0

            except (InvalidGitRepositoryError, NoSuchPathError):
                new_parent = parent_directory.parent
                # If we encountered root and cannot move further: no .git directory located in the entire path
                if new_parent == parent_directory:
                    return True
                parent_directory = new_parent

    @staticmethod
    def is_find_by_ext_file(config: Config, extension: str) -> bool:
        """
        Checks whether file has suspicious extension

        Args:
            config: Config
            extension: str - may be only file name with extension

        Return:
            True when the feature is configured and the file extension matches
        """
        return config.find_by_ext and extension in config.find_by_ext_list

    @staticmethod
    def check_exclude_file(config: Config, path: str) -> bool:
        """
        Checks whether file should be excluded

        Args:
            config: Config
            path: str - full path preferred

        Return:
            True when the file full path should be excluded according config
        """
        path = path.replace('\\', '/')
        lower_path = path.lower()
        if config.not_allowed_path_pattern.match(lower_path):
            return True
        for exclude_pattern in config.exclude_patterns:
            if exclude_pattern.match(lower_path):
                return True
        for exclude_path in config.exclude_paths:
            # must be case-sensitive
            if exclude_path in path:
                return True
        file_extension = Util.get_extension(lower_path, lower=False)
        if file_extension in config.exclude_extensions:
            return True
        if not config.depth and file_extension in config.exclude_containers:
            return True
        # --depth or --doc enables scan for all documents extensions
        if not (config.depth or config.doc) and file_extension in config.exclude_documents:
            return True
        return False

    @staticmethod
    def check_file_size(config: Config, reference: Union[str, Path, io.BytesIO, Tuple[Union[str, Path],
                                                                                      io.BytesIO]]) -> bool:
        """
        Checks whether the file is over the size limit from configuration or less MIN_DATA_LEN

        Args:
            config: Config
            reference: various types of a file reference

        Return:
            True when the file is oversize or less than MIN_DATA_LEN, or unsupported
        """
        path = reference[1] if isinstance(reference, tuple) else reference
        if isinstance(path, str) or isinstance(path, Path):
            file_size = os.path.getsize(path)
        elif isinstance(path, io.BytesIO):
            current_pos = path.tell()
            path.seek(0, io.SEEK_END)
            file_size = path.tell() - current_pos
            path.seek(current_pos, io.SEEK_SET)
        else:
            logger.error(f"Unknown path type: {path}")
            return True

        if MIN_DATA_LEN > file_size:
            logger.debug(f"Size ({file_size}) of the file '{path}' is too small")
            return True
        elif isinstance(config.size_limit, int) and config.size_limit < file_size:
            logger.warning(f"Size ({file_size}) of the file '{path}' is over limit ({config.size_limit})")
            return True

        return False



import contextlib
import logging
import re
import string
from typing import List

from credsweeper.common.constants import PEM_BEGIN_PATTERN, PEM_END_PATTERN, Chars
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.utils import Util
from credsweeper.utils.entropy_validator import EntropyValidator

logger = logging.getLogger(__name__)


class PemKeyDetector:
    """Class to detect PEM PRIVATE keys only"""
    base64set = set(string.ascii_uppercase) | set(string.ascii_lowercase) | set(string.digits) | {'+', '/', '='}

    ignore_starts = [PEM_BEGIN_PATTERN, "Proc-Type", "Version", "DEK-Info"]
    wrap_characters = "\\'\";,[]#*!"
    remove_characters = string.whitespace + wrap_characters
    # last line contains 4 symbols, at least
    re_pem_begin = re.compile(r"(?P<value>" + PEM_BEGIN_PATTERN + r"\s(?!ENCRYPTED)[^-]*PRIVATE[^-]*KEY[^-]*-----"
                              r"(.+" + PEM_END_PATTERN + r"[^-]+KEY[^-]*-----)?)")
    re_value_pem = re.compile(r"(?P<value>([^-]*" + PEM_END_PATTERN +
                              r"[^-]+-----)|(([a-zA-Z0-9/+=]{64}.*)?[a-zA-Z0-9/+=]{4})+)")

    @classmethod
    def detect_pem_key(cls, config: Config, target: AnalysisTarget) -> List[LineData]:
        """Detects PEM key in single line and with iterative for next lines according
        https://www.rfc-editor.org/rfc/rfc7468

        Args:
            config: Config
            target: Analysis target

        Return:
            List of LineData with found PEM

        """
        line_data: List[LineData] = []
        key_data = ""
        # get line with -----BEGIN which may contain full key
        first_line = LineData(config, target.line, target.line_pos, target.line_num, target.file_path, target.file_type,
                              target.info, cls.re_pem_begin)
        line_data.append(first_line)
        # protection check for case when first line starts from 0
        start_pos = target.line_pos if 0 <= target.line_pos else 0
        finish_pos = min(start_pos + 200, target.lines_len)
        begin_pattern_not_passed = True
        for line_pos in range(start_pos, finish_pos):
            line = target.lines[line_pos]
            if target.line_pos != line_pos:
                _line = LineData(config, line, line_pos, target.line_nums[line_pos], target.file_path, target.file_type,
                                 target.info, cls.re_value_pem)
                line_data.append(_line)
            # replace escaped line ends with real and process them - PEM does not contain '\' sign
            while "\\\\" in line:
                line = line.replace("\\\\", "\\")
            sublines = line.replace("\\r", '\n').replace("\\n", '\n').splitlines()
            for subline in sublines:
                if begin_pattern_not_passed or cls.is_leading_config_line(subline):
                    if PEM_BEGIN_PATTERN in subline:
                        begin_pattern_not_passed = False
                    continue
                elif PEM_END_PATTERN in subline:
                    if "PGP" in target.line_strip:
                        # Check if entropy is high enough for base64 set with padding sign
                        entropy_validator = EntropyValidator(key_data, Chars.BASE64STDPAD_CHARS)
                        if entropy_validator.valid:
                            return line_data
                        logger.debug("Filtered with entropy %f '%s'", entropy_validator.entropy, key_data)
                    if "OPENSSH" in target.line_strip:
                        # Check whether the key is encrypted
                        with contextlib.suppress(Exception):
                            decoded = Util.decode_base64(key_data, urlsafe_detect=True)
                            if 32 < len(decoded) and b"bcrypt" not in decoded:
                                # 256 bits is the minimal size of Ed25519 keys
                                # all OK - the key is not encrypted in this top level
                                return line_data
                        logger.debug("Filtered with size or bcrypt '%s'", key_data)
                    else:
                        with contextlib.suppress(Exception):
                            decoded = Util.decode_base64(key_data, urlsafe_detect=True)
                            if Util.is_asn1(decoded):
                                # all OK - the key is not encrypted in this top level
                                return line_data
                        logger.debug("Filtered with non asn1 '%s'", key_data)
                    return []
                else:
                    sanitized_line = cls.sanitize_line(subline)
                    # PEM key line should not contain spaces or . (and especially not ...)
                    for i in sanitized_line:
                        if i not in cls.base64set:
                            return []
                    key_data += sanitized_line
        return []

    @classmethod
    def sanitize_line(cls, line: str, recurse_level: int = 5) -> str:
        """Remove common symbols that can surround PEM keys inside code.

        Examples::

            `# ZZAWarrA1`
            `* ZZAWarrA1`
            `  "ZZAWarrA1\\n" + `

        Args:
            line: Line to be cleaned
            recurse_level: to avoid infinite loop in case when removed symbol inside base64 encoded

        Return:
            line with special characters removed from both ends

        """
        recurse_level -= 1

        if 0 > recurse_level:
            return line

        # Note that this strip would remove `\n` but not `\\n`
        line = line.strip(string.whitespace)
        if line.startswith("//"):
            # simplify first condition for speed-up of doxygen style processing
            if line.startswith("// ") or line.startswith("/// "):
                # Assume that the commented line is to be separated from base64 code, it may be a part of PEM, otherwise
                line = line[3:]
        if line.startswith("/*"):
            line = line[2:]
        if line.endswith("*/"):
            line = line[:-2]
        if line.endswith("\\"):
            # line carry in many languages
            line = line[:-1]

        # remove concatenation carefully only when it is not part of base64
        if line.startswith('+') and 1 < len(line) and line[1] not in cls.base64set:
            line = line[1:]
        if line.endswith('+') and 2 < len(line) and line[-2] not in cls.base64set:
            line = line[:-1]

        line = line.strip(cls.remove_characters)
        # check whether new iteration requires
        for x in string.whitespace:
            if line.startswith(x) or line.endswith(x):
                return cls.sanitize_line(line, recurse_level)

        for x in cls.wrap_characters:
            if x in line:
                return cls.sanitize_line(line, recurse_level)

        return line

    @classmethod
    def is_leading_config_line(cls, line: str) -> bool:
        """Remove non-key lines from the beginning of a list.

        Example lines with non-key leading lines:

        .. code-block:: text

            Proc-Type: 4,ENCRYPTED
            DEK-Info: DEK-Info: AES-256-CBC,2AA219GG746F88F6DDA0D852A0FD3211

            ZZAWarrA1...

        Args:
            line: Line to be checked

        Return:
            True if the line is not a part of encoded data but leading config

        """
        if 0 == len(line):
            return True
        for ignore_string in cls.ignore_starts:
            if ignore_string in line:
                return True
        return False



from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueSimilarityCheck(Filter):
    """Check if candidate value is at least 70% same as candidate keyword. Like: `secret = "mysecret"`."""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        # Cannot evaluate if key is None
        if line_data.key is None:
            return False
        if line_data.key.lower() in line_data.value.lower() and \
                len(line_data.key) / len(line_data.value) >= 0.7:
            return True
        if line_data.variable is not None and line_data.value in line_data.variable:
            return True
        return False



from abc import abstractmethod, ABC
from typing import List

from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider
from credsweeper.scanner import Scanner


class AbstractScanner(ABC):
    """Base abstract class for all recursive scanners"""

    @property
    @abstractmethod
    def config(self) -> Config:
        """Abstract property to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @property
    @abstractmethod
    def scanner(self) -> Scanner:
        """Abstract property to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @abstractmethod
    def recursive_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int = 0,  #
            recursive_limit_size: int = 0) -> List[Candidate]:
        """Abstract method to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @abstractmethod
    def structure_scan(
            self,  #
            struct_provider: StructContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> List[Candidate]:
        """Abstract method to be defined in DeepScanner"""
        raise NotImplementedError(__name__)



import base64
import contextlib
import string

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueBase32DataCheck(Filter):
    """
    Check that candidate is NOT an ascii encoded string with entropy check
    """

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received weird base32 token which must be a random string

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        value = line_data.value
        # check whether digits and upper cases present
        for string_set in [string.digits, string.ascii_uppercase]:
            for digit in string_set:
                if digit in value:
                    break
            else:
                return True
        # check whether decoded bytes have enough entropy
        with contextlib.suppress(Exception):
            decoded = base64.b32decode(value)
            return Util.is_ascii_entropy_validate(decoded)
        return True



import re

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature


class SearchInAttribute(Feature):
    """Abstract feature returns boolean for matched pattern in member of first LineData"""

    def __init__(self, pattern: str, attribute: str):
        super().__init__()
        self.pattern = re.compile(pattern)
        self.attribute = attribute

    def extract(self, candidate: Candidate) -> bool:
        """Returns boolean for first LineData member"""
        if attribute := getattr(candidate.line_data_list[0], self.attribute, None):
            if self.pattern.search(attribute):
                return True
        return False



import contextlib
import string

from cryptography.hazmat.primitives import serialization

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueBase64KeyCheck(Filter):
    """Check that candidate contains base64 encoded private key"""

    def __init__(self, config: Config = None) -> None:
        self.config = config

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """

        with contextlib.suppress(Exception):
            text = line_data.value
            # replace to space any escaped sequence except space from string.whitespace
            for x in ["\\t", "\\n", "\\r", "\\v", "\\f"]:
                text = text.replace(x, ' ')
            for x in string.whitespace:
                text = text.replace(x, '')
            # clean sequence concatenation case:
            text = text.replace("'+'", '')
            text = text.replace('"+"', '')
            # possibly url based escaping:
            text = text.replace('%2B', '+')
            text = text.replace('%2F', '/')
            text = text.replace('%3D', '=')
            # clean any other chars which should not appear
            for x in ["'", '"', '\\', ',']:
                text = text.replace(x, "")
            # only PEM standard encoding supported in regex pattern to cut off ending of the key
            key = Util.decode_base64(text, padding_safe=True, urlsafe_detect=False)
            private_key = serialization.load_der_private_key(key, password=None)
            if 0 < private_key.key_size:  # type: ignore
                # access to size field check - some types have no size
                return False
        return True



import contextlib
import json

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueGrafanaCheck(Filter):
    """Grafana Provisioned API Key and Access Policy Token"""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        with contextlib.suppress(Exception):
            if line_data.value.startswith("glc_"):
                # Grafana Access Policy Token
                decoded = Util.decode_base64(line_data.value[4:], padding_safe=True, urlsafe_detect=True)
                keys = ["o", "n", "k", "m"]
            else:
                # Grafana Provisioned API Key
                decoded = Util.decode_base64(line_data.value, padding_safe=True, urlsafe_detect=True)
                keys = ["n", "k", "id"]
            if payload := json.loads(decoded):
                for key in keys:
                    if key not in payload:
                        return True
                return False
        return True



from typing import Dict, List, Tuple

from credsweeper.credentials.candidate import Candidate
from credsweeper.credentials.candidate_key import CandidateKey


class CandidateGroupGenerator:
    """CandidateGroupGenerator"""

    def __init__(self) -> None:
        self.grouped_candidates: Dict[CandidateKey, List[Candidate]] = {}

    @property
    def grouped_candidates(self) -> Dict[CandidateKey, List[Candidate]]:
        """property getter"""
        return self._grouped_candidates

    @grouped_candidates.setter
    def grouped_candidates(self, grouped_candidates: Dict[CandidateKey, List[Candidate]]) -> None:
        """property setter"""
        self._grouped_candidates = grouped_candidates

    def __contains__(self, key: CandidateKey) -> bool:
        return key in self.grouped_candidates

    def __getitem__(self, key) -> List[Candidate]:
        return self.grouped_candidates[key]

    def __setitem__(self, key: CandidateKey, value: List[Candidate]) -> None:
        self.grouped_candidates[key] = value

    def __len__(self) -> int:
        return len(self.grouped_candidates)

    def items(self) -> List[Tuple[CandidateKey, List[Candidate]]]:
        """getter"""
        return list(self.grouped_candidates.items())



import math
from typing import Dict, List, Set

import numpy as np

from credsweeper.common.constants import Chars, ML_HUNK
from credsweeper.credentials import Candidate
from credsweeper.file_handler.data_content_provider import MIN_DATA_LEN
from credsweeper.ml_model.features.feature import Feature


class EntropyEvaluation(Feature):
    """
    Renyi, Shannon entropy evaluation with Hartley entropy normalization.
    Augmentation with possible set of chars (hex, base64, etc.)
    Analyse only begin of the value

    See next link for details:
    https://digitalassets.lib.berkeley.edu/math/ucb/text/math_s4_v1_article-27.pdf

    """

    def __init__(self) -> None:
        """Class initializer"""
        super().__init__()
        # Max size of ML analyzed value is ML_HUNK but value may be bigger
        self.hunk_size = 4 * ML_HUNK
        self.log2_cache: Dict[int, float] = {x: math.log2(x) for x in range(4, self.hunk_size + 1)}
        self.char_sets: List[Set[str]] = [set(x.value) for x in Chars]

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns real entropy and possible sets of characters"""
        # only head of value will be analyzed
        result: np.ndarray = np.zeros(shape=3 + len(self.char_sets), dtype=np.float32)
        value = candidate.line_data_list[0].value[:self.hunk_size]
        size = len(value)
        uniq, counts = np.unique(list(value), return_counts=True)
        if MIN_DATA_LEN <= size:
            # evaluate the entropy for a value of at least 4
            probabilities = counts / size
            hartley_entropy = self.log2_cache.get(size, -1.0)
            assert hartley_entropy, str(candidate)

            # renyi_entropy alpha=0.5
            sum_prob_05 = np.sum(probabilities**0.5)
            renyi_entropy_05 = 2 * np.log2(sum_prob_05)
            result[0] = renyi_entropy_05 / hartley_entropy

            # shannon_entropy or renyi_entropy alpha=1
            shannon_entropy = -np.sum(probabilities * np.log2(probabilities))
            result[1] = shannon_entropy / hartley_entropy

            # renyi_entropy alpha=2
            sum_prob_2 = np.sum(probabilities**2)
            renyi_entropy_2 = -1 * np.log2(sum_prob_2)
            result[2] = renyi_entropy_2 / hartley_entropy

        if 0 < size:
            # check charset for non-zero value
            # use the new variable to deal with mypy
            uniq_set = set(uniq)
            for n, i in enumerate(self.char_sets, start=3):
                if not uniq_set.difference(i):
                    result[n] = 1.0

        return result



from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature


class IsSecretNumeric(Feature):
    """Feature is true if candidate value is a numerical value."""

    def extract(self, candidate: Candidate) -> bool:
        try:
            float(candidate.line_data_list[0].value)
            return True
        except ValueError:
            return False



import bz2
import logging
from abc import ABC
from pathlib import Path
from typing import List, Optional

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class Bzip2Scanner(AbstractScanner, ABC):
    """Implements bzip2 scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data from bzip2 archive and launches data_scan"""
        try:
            file_path = Path(data_provider.file_path)
            new_path = file_path.as_posix()
            if ".bz2" == file_path.suffix:
                new_path = new_path[:-4]
            bzip2_content_provider = DataContentProvider(data=bz2.decompress(data_provider.data),
                                                         file_path=new_path,
                                                         file_type=Util.get_extension(new_path),
                                                         info=f"{data_provider.info}|BZIP2:{new_path}")
            new_limit = recursive_limit_size - len(bzip2_content_provider.data)
            bzip2_candidates = self.recursive_scan(bzip2_content_provider, depth, new_limit)
            return bzip2_candidates
        except Exception as bzip2_exc:
            logger.error(f"{data_provider.file_path}:{bzip2_exc}")
        return None



from functools import cached_property
from typing import List, Optional

from credsweeper.file_handler.descriptor import Descriptor


class AnalysisTarget:
    """AnalysisTarget"""

    def __init__(
        self,
        line_pos: int,
        lines: List[str],
        line_nums: List[int],
        descriptor: Descriptor,
        line: Optional[str] = None,
        offset: Optional[int] = None,
    ):
        self.__line_pos = line_pos
        self.__lines = lines
        self.__line_nums = line_nums
        self.__descriptor = descriptor
        self.__line = line
        self.__offset = offset

    @cached_property
    def offset(self) -> Optional[int]:
        """cached value"""
        # when the offset is not None - it means that original line was split into chunks
        return self.__offset

    @cached_property
    def line(self) -> str:
        """cached value"""
        if self.__line is None:
            # normal target
            return self.__lines[self.__line_pos]
        else:
            # chunked target
            return self.__line

    @cached_property
    def line_len(self) -> int:
        """cached value"""
        return len(self.line)

    @cached_property
    def line_strip(self) -> str:
        """cached value"""
        return self.line.strip()

    @cached_property
    def line_strip_len(self) -> int:
        """cached value"""
        return len(self.line_strip)

    @cached_property
    def line_lower(self) -> str:
        """cached value"""
        return self.line.lower()

    @cached_property
    def line_lower_strip(self) -> str:
        """cached value"""
        return self.line_lower.strip()

    @cached_property
    def lines(self) -> List[str]:
        """cached value"""
        return self.__lines

    @cached_property
    def lines_len(self) -> int:
        """cached value"""
        return len(self.__lines)

    @cached_property
    def line_pos(self) -> int:
        """cached value"""
        return self.__line_pos

    @cached_property
    def line_num(self) -> int:
        """cached value"""
        return self.__line_nums[self.__line_pos]

    @cached_property
    def line_nums(self) -> List[int]:
        """cached value"""
        return self.__line_nums

    @cached_property
    def file_path(self) -> Optional[str]:
        """cached value"""
        return self.__descriptor.path

    @cached_property
    def file_type(self) -> Optional[str]:
        """cached value"""
        return self.__descriptor.extension

    @cached_property
    def info(self) -> Optional[str]:
        """cached value"""
        return self.__descriptor.info

    @cached_property
    def descriptor(self) -> Descriptor:
        """cached value"""
        return self.__descriptor



from typing import List

import numpy as np

from credsweeper.common.constants import CHUNK_SIZE
from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.word_in import WordIn
from credsweeper.utils import Util


class WordInLine(WordIn):
    """Feature is true if line contains at least one word from predefined list."""

    def __init__(self, words: List[str]) -> None:
        """Feature returns array of matching words

        Args:
            words: list of predefined words - MUST BE IN LOWER CASE

        """
        super().__init__(words)

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns true if any words in first line"""
        subtext = Util.subtext(candidate.line_data_list[0].line, candidate.line_data_list[0].value_start, CHUNK_SIZE)
        if subtext:
            return self.word_in_str(subtext.lower())
        else:
            return np.array([np.zeros(shape=[self.dimension], dtype=np.int8)])



import io
import logging
from abc import ABC
from typing import List, Optional
from zipfile import ZipFile

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class ZipScanner(AbstractScanner, ABC):
    """Implements zip scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts files one by one from zip archives and launches data_scan"""
        try:
            candidates = []
            with ZipFile(io.BytesIO(data_provider.data)) as zf:
                for zfl in zf.infolist():
                    # skip directory
                    if zfl.is_dir():
                        continue
                    if FilePathExtractor.check_exclude_file(self.config, zfl.filename):
                        continue
                    if 0 > recursive_limit_size - zfl.file_size:
                        logger.error(f"{zfl.filename}: size {zfl.file_size}"
                                     f" is over limit {recursive_limit_size} depth:{depth}")
                        continue
                    with zf.open(zfl) as f:
                        zip_content_provider = DataContentProvider(data=f.read(),
                                                                   file_path=data_provider.file_path,
                                                                   file_type=Util.get_extension(zfl.filename),
                                                                   info=f"{data_provider.info}|ZIP:{zfl.filename}")
                        # nevertheless use extracted data size
                        new_limit = recursive_limit_size - len(zip_content_provider.data)
                        zip_candidates = self.recursive_scan(zip_content_provider, depth, new_limit)
                        candidates.extend(zip_candidates)
            return candidates
        except Exception as zip_exc:
            # too many exception types might be produced with broken zip
            logger.error(f"{data_provider.file_path}:{zip_exc}")
        return None



from credsweeper.common import static_keyword_checklist
from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature


class MorphemeDense(Feature):
    """Feature calculates morphemes density for a value"""

    def __init__(self) -> None:
        """Class initializer"""
        super().__init__()

    def extract(self, candidate: Candidate) -> float:
        if value := candidate.line_data_list[0].value.lower():
            morphemes_counter = 0
            for morpheme in static_keyword_checklist.morpheme_set:
                if morpheme in value:
                    morphemes_counter += 1
            # normalization: minimal morpheme length is 3
            return 3.0 * morphemes_counter / len(value)
        else:
            # empty value case
            return 0.0



from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueDictionaryValueLengthCheck(Filter):
    """Check that candidate length is between 5 and 30."""

    def __init__(self, config: Config = None, min_len: int = 4, max_len: int = 31) -> None:
        self.min_len = min_len
        self.max_len = max_len

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if self.min_len <= len(line_data.value) <= self.max_len:
            return False
        else:
            return True



import logging
from abc import ABC
from typing import List, Optional

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class HtmlScanner(AbstractScanner, ABC):
    """Implements html scanning if possible"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to represent data as html text and scan as text lines"""
        if data_provider.represent_as_html(depth, recursive_limit_size,
                                           self.scanner.keywords_required_substrings_check):
            string_data_provider = StringContentProvider(lines=data_provider.lines,
                                                         line_numbers=data_provider.line_numbers,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|HTML")
            return self.scanner.scan(string_data_provider)
        return None



from credsweeper.common import static_keyword_checklist
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueCoupleKeywordCheck(Filter):
    """Check value if TWO words from morphemes checklist exists in value"""

    def __init__(self, config: Config = None, threshold=1) -> None:
        # threshold - minimum morphemes number in a value
        self.threshold = threshold

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        return static_keyword_checklist.check_morphemes(line_data.value.lower(), self.threshold)



import gzip
import io
import logging
from abc import ABC
from pathlib import Path
from typing import List, Optional

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class GzipScanner(AbstractScanner, ABC):
    """Realises gzip scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data from gzip archive and launches data_scan"""
        try:
            with gzip.open(io.BytesIO(data_provider.data)) as f:
                file_path = Path(data_provider.file_path)
                new_path = file_path.as_posix()
                if ".gz" == file_path.suffix:
                    new_path = new_path[:-3]
                gzip_content_provider = DataContentProvider(data=f.read(),
                                                            file_path=new_path,
                                                            file_type=Util.get_extension(new_path),
                                                            info=f"{data_provider.info}|GZIP:{new_path}")
                new_limit = recursive_limit_size - len(gzip_content_provider.data)
                gzip_candidates = self.recursive_scan(gzip_content_provider, depth, new_limit)
                return gzip_candidates
        except Exception as gzip_exc:
            logger.error(f"{data_provider.file_path}:{gzip_exc}")
        return None



import contextlib
import re
import statistics

from credsweeper.common.constants import Chars
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.filters.value_entropy_base64_check import ValueEntropyBase64Check
from credsweeper.utils import Util


class ValueBase64PartCheck(Filter):
    """
    Check that candidate is NOT a part of base64 long line
    """

    base64_pattern = re.compile(r"^(\\{1,8}[0abfnrtv]|[0-9A-Za-z+/=]){1,4000}")
    base64_set = set(Chars.BASE64STDPAD_CHARS.value)

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received weird base64 token which must be a random string

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """

        with contextlib.suppress(Exception):
            line = line_data.line
            len_line = len(line)
            value = line_data.value
            len_value = len(value)
            if 0 == line_data.value_start and len_line >= 2 * len_value \
                    or 0 < line_data.value_start and line[line_data.value_start - 1] in ('/', '+', '\\', '%') \
                    or 0 < line_data.value_end < len_line and line[line_data.value_end] in ('/', '+', '\\', '%'):

                if '-' in value or '_' in value:
                    # the value contains url-safe chars, so '/' or '+' is a delimiter
                    return False

                left_start = line_data.value_start - len_value
                if 0 > left_start:
                    left_start = 0
                right_end = line_data.value_end + len_value
                if len_line < right_end:
                    right_end = len_line

                hunk_size = right_end - left_start

                if hunk_size == 3 * len_value:
                    # simple analysis for maximal data size
                    if self.base64_pattern.match(line[left_start:right_end]):
                        # obvious case: all characters are base64 standard
                        return True
                elif right_end - left_start >= 2 * len_value:
                    # simple analysis for data too large to yield sensible insights
                    part_set = set(line[left_start:right_end])
                    if not part_set.difference(self.base64_set):
                        # obvious case: all characters are base64 standard
                        return True

                left_part = line[left_start:line_data.value_start]
                len_left = len(left_part)
                right_part = line[line_data.value_end:right_end]
                len_right = len(right_part)

                min_entropy_value = ValueEntropyBase64Check.get_min_data_entropy(len_value)
                value_entropy = Util.get_shannon_entropy(value, Chars.BASE64STD_CHARS.value)

                if ValueEntropyBase64Check.min_length < len_left:
                    left_entropy = Util.get_shannon_entropy(left_part, Chars.BASE64STD_CHARS.value)
                    if len_left < len_value:
                        left_entropy *= len_value / len_left
                else:
                    left_entropy = min_entropy_value

                if ValueEntropyBase64Check.min_length < len_right:
                    right_entropy = Util.get_shannon_entropy(right_part, Chars.BASE64STD_CHARS.value)
                    if len_right < len_value:
                        left_entropy *= len_right / len_left
                else:
                    right_entropy = min_entropy_value

                data = [left_entropy, value_entropy, right_entropy, min_entropy_value]
                avg = statistics.mean(data)
                stdev = statistics.stdev(data, avg)
                avg_min = avg - 1.1 * stdev
                if avg_min <= left_entropy and avg_min <= right_entropy:
                    # high entropy of bound parts looks like a part of base64 long line
                    return True

        return False



from typing import List

from credsweeper.common.constants import RuleType
from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.rules import Rule
from credsweeper.scanner.scan_type import ScanType


class MultiPattern(ScanType):
    """Check if line is a part of a multi-line credential and second part is present within MAX_SEARCH_MARGIN lines.

    Parameters:
        MAX_SEARCH_MARGIN: Int constant. Number of lines around current to perform search for the second part

    """

    MAX_SEARCH_MARGIN = 10

    @classmethod
    def run(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Check if multiline credential present if the file within MAX_SEARCH_MARGIN range from current line_num.

        Args:
            config: user configs
            rule: Rule object to check current line. Should be a multi-pattern rule
            target: Analysis target

        Return:
            List of Candidates if pattern defined in a rule is present in a line
            and second part of multi-pattern rule is present within MAX_SEARCH_MARGIN from the line.
            Empty list (False) - otherwise.

        """
        assert rule.rule_type == RuleType.MULTI, \
            "Rules provided to MultiPattern.run should have pattern_type equal to MULTI_PATTERN"

        candidates = cls._get_candidates(config, rule, target)
        if not candidates:
            return candidates
        for candidate in candidates:
            line_pos_margin = 1
            while line_pos_margin <= cls.MAX_SEARCH_MARGIN:
                candi_line_pos_backward = candidate.line_data_list[0].line_pos - line_pos_margin
                if 0 <= candi_line_pos_backward < target.lines_len:
                    if cls._scan(config, candidate, candi_line_pos_backward, target, rule):
                        break
                candi_line_pos_forward = candidate.line_data_list[0].line_pos + line_pos_margin
                if candi_line_pos_forward < target.lines_len:
                    if cls._scan(config, candidate, candi_line_pos_forward, target, rule):
                        break
                line_pos_margin += 1

            # Check if found multi line
            if len(candidate.line_data_list) == 1:
                if not cls._scan(config, candidate, candidate.line_data_list[0].line_pos, target, rule):
                    # last resort - to find the credential in the same line
                    return []

        return candidates

    @classmethod
    def _scan(cls, config: Config, candidate: Candidate, candi_line_pos: int, target: AnalysisTarget,
              rule: Rule) -> bool:
        """Search for second part of multiline rule near the current line.

        Automatically update candidate with detected line if any.

        Args:
            config: dict, scanner configuration
            candidate: Current credential candidate detected in the line
            candi_line_pos: line position of lines around candidate to perform search
            target: Analysis target
            rule: Rule object to check current line. Should be a multi-pattern rule

        Return:
            Boolean. True if second part detected. False otherwise

        """
        new_target = AnalysisTarget(candi_line_pos, target.lines, target.line_nums, target.descriptor)

        line_data_list = cls.get_line_data_list(config=config,
                                                target=new_target,
                                                pattern=rule.patterns[1],
                                                filters=rule.filters)

        if not line_data_list:
            return False
        else:
            candidate.line_data_list.extend(line_data_list)
            return True



import numpy as np

from credsweeper.common.constants import ML_HUNK
from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature


class LengthOfAttribute(Feature):
    """Abstract class for obtain a normalized value of length with max size of hunk"""

    def __init__(self, attribute: str):
        super().__init__()
        if "line" == attribute:
            self.hunk_plus = 2 * ML_HUNK + 1
        elif "value" == attribute or "variable" == attribute:
            self.hunk_plus = ML_HUNK + 1
        else:
            raise ValueError(f"Not supported attribute '{attribute}'")
        self.attribute = attribute

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns boolean for first LineData member"""
        if attribute := getattr(candidate.line_data_list[0], self.attribute, None):
            if len(attribute) < self.hunk_plus:
                # should be in (0, 1)
                return np.array([len(attribute) / self.hunk_plus])
            else:
                # 1.0 means the attribute is oversize
                return np.array([1.0])
        # the attribute is empty
        return np.array([0.0])



import logging
from abc import ABC
from typing import List, Optional

from bs4 import BeautifulSoup
from lxml import etree

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class MxfileScanner(AbstractScanner, ABC):
    """Scanner for drawio diagram"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to get text data from the xml format"""
        try:
            lines = []
            line_numbers = []
            tree = etree.fromstring(data_provider.text)
            for element in tree.iter():
                if "mxCell" == getattr(element, "tag"):
                    line_number = element.sourceline
                    attr = getattr(element, "attrib")
                    if attr is None or not (value := attr.get("value")):
                        continue
                    if html := BeautifulSoup(value, features="html.parser"):
                        _, value_lines, __ = data_provider.simple_html_representation(html)
                        for line in value_lines:
                            lines.append(line)
                            line_numbers.append(line_number)
            mxfile_data_provider = StringContentProvider(lines=lines,
                                                         line_numbers=line_numbers,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|MXFILE")
            return self.scanner.scan(mxfile_data_provider)
        except Exception as exc:
            logger.error(exc)
        return None



import string
import typing
from enum import Enum
from typing import Optional, Union


class Severity(Enum):
    """Severity of candidate"""
    CRITICAL = "critical"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    INFO = "info"

    def __lt__(self, other) -> bool:
        if self == Severity.INFO:
            return other is not Severity.INFO
        elif self == Severity.LOW:
            return other in [Severity.MEDIUM, Severity.HIGH, Severity.CRITICAL]
        elif self == Severity.MEDIUM:
            return other in [Severity.HIGH, Severity.CRITICAL]
        elif self == Severity.HIGH:
            return other is Severity.CRITICAL
        return False

    @staticmethod
    def get(severity: Union[str, "Severity"]) -> Optional["Severity"]:
        """returns Severity value from string or None"""
        if isinstance(severity, Severity):
            return severity
        if isinstance(severity, str):
            value = getattr(Severity, severity.strip().upper(), None)
            if isinstance(value, Severity):
                return value
        return None


class Confidence(Enum):
    """Confidence of candidate"""
    STRONG = "strong"
    MODERATE = "moderate"
    WEAK = "weak"

    def __lt__(self, other) -> bool:
        if self == Confidence.WEAK:
            return other is not Confidence.WEAK
        elif self == Confidence.MODERATE:
            return other is Confidence.STRONG
        return False

    @staticmethod
    def get(confidence: Union[str, "Confidence"]) -> Optional["Confidence"]:
        """returns Confidence value from string or None"""
        if isinstance(confidence, Confidence):
            return confidence
        if isinstance(confidence, str):
            value = getattr(Confidence, confidence.strip().upper(), None)
            if isinstance(value, Confidence):
                return value
        return None


BASE64COMMON = string.ascii_uppercase + string.ascii_lowercase + string.digits


class Chars(Enum):
    """Stores enumeration of characters sets of encoding dictionaries"""

    # set of characters, hexadecimal numeral system (Base16). Upper- and lowercase
    HEX_CHARS = string.digits + "ABCDEFabcdef"
    # UUID charset in uppercase
    UUID_UPPER_CHARS = string.digits + "ABCDEF-"
    # UUID charset in lowercase
    UUID_LOWER_CHARS = string.digits + "abcdef-"
    # set of characters, hexadecimal numeral system (Base16). Uppercase
    BASE16UPPER = string.digits + "ABCDEF"
    # set of characters, hexadecimal numeral system (Base16). Lowercase
    BASE16LOWER = string.digits + "abcdef"
    # set of 32 characters, used in Base32 encoding
    BASE32_CHARS = string.ascii_uppercase + "234567"
    # set of 36 characters, used in Base36 encoding
    BASE36_CHARS = string.digits + string.ascii_lowercase
    # base62 set https://en.wikipedia.org/wiki/Base62
    BASE62_CHARS = string.digits + string.ascii_uppercase + string.ascii_lowercase
    # URL- and filename-safe standard
    BASE64URL_CHARS = BASE64COMMON + "-_"
    # URL- and filename-safe standard plus padding sign
    BASE64URLPAD_CHARS = BASE64COMMON + "-_="
    # standard base64 charset
    BASE64STD_CHARS = BASE64COMMON + "+/"
    # standard base64 plus padding sign
    BASE64STDPAD_CHARS = BASE64COMMON + "+/="
    # except whitespaces
    ASCII_VISIBLE = string.digits + string.ascii_letters + string.punctuation
    # all printable symbols
    ASCII_PRINTABLE = string.printable


ENTROPY_LIMIT_BASE64 = 4.5
ENTROPY_LIMIT_BASE3x = 3


class GroupType(Enum):
    """Group type - used in Group constructor for load predefined set of filters"""
    KEYWORD = "keyword"
    PATTERN = "pattern"
    # for empty filter set
    DEFAULT = "default"


class RuleType(Enum):
    """Rule type"""
    # combine pattern with predefined structure
    KEYWORD = "keyword"
    # use patterns as-is. all patterns must be found in target (line)
    PATTERN = "pattern"
    # single value to detect pem format with specific scanner
    PEM_KEY = "pem_key"
    # When first pattern found - second will be searched in adjoining lines
    MULTI = "multi"


class ThresholdPreset(Enum):
    """Preset threshold to simplify precision/recall selection for the user."""
    lowest = "lowest"
    low = "low"
    medium = "medium"
    high = "high"
    highest = "highest"


class DiffRowType(Enum):
    """Diff type of row"""
    ADDED = "added"
    DELETED = "deleted"


StartEnd = typing.NamedTuple("StartEnd", [("start", int), ("end", int)])

MIN_VARIABLE_LENGTH = 1
MIN_SEPARATOR_LENGTH = 1
MIN_VALUE_LENGTH = 4
# if the line is oversize - it will be scanned by chunks with overlapping
MAX_LINE_LENGTH = 8000
# the size for overlapping chunks must be less than MAX_LINE_LENGTH
CHUNK_SIZE = 4000
OVERLAP_SIZE = 1000
CHUNK_STEP_SIZE = CHUNK_SIZE - OVERLAP_SIZE
# ML hunk size to limit of variable or value size and get substring near value
ML_HUNK = 80
""" values according https://docs.python.org/3/library/codecs.html """
UTF_8 = "utf_8"
UTF_16 = "utf_16"
LATIN_1 = "latin_1"
ASCII = "ascii"

DEFAULT_ENCODING = UTF_8

# LATIN_1 has to be placed at end to apply binary file detection
AVAILABLE_ENCODINGS = [UTF_8, UTF_16, LATIN_1]

# to limit memory usage in case of recursive scan
RECURSIVE_SCAN_LIMITATION = 1 << 30

# default value for config and ValuePatternCheck
DEFAULT_PATTERN_LEN = 4

# PEM x509 patterns
PEM_BEGIN_PATTERN = "-----BEGIN"
PEM_END_PATTERN = "-----END"

# similar min_line_len in rule_template - no real credential in data less than 8 bytes
MIN_DATA_LEN = 8



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters import ValueCoupleKeywordCheck, ValueCamelCaseCheck, ValueNumberCheck, ValuePatternCheck
from credsweeper.filters.group import Group


class TokenPattern(Group):
    """Token Pattern"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueCoupleKeywordCheck(),
            ValueNumberCheck(),
            ValueCamelCaseCheck(),
            ValuePatternCheck(pattern_len=config.pattern_len)
        ]



import datetime
import logging
from typing import List, Optional, Any, Tuple, Union

from credsweeper.common.constants import RECURSIVE_SCAN_LIMITATION
from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.credentials.augment_candidates import augment_candidates
from credsweeper.file_handler.byte_content_provider import ByteContentProvider
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.file_handler.text_content_provider import TextContentProvider
from credsweeper.scanner import Scanner
from credsweeper.utils import Util
from .byte_scanner import ByteScanner
from .bzip2_scanner import Bzip2Scanner
from .docx_scanner import DocxScanner
from .eml_scanner import EmlScanner
from .encoder_scanner import EncoderScanner
from .gzip_scanner import GzipScanner
from .html_scanner import HtmlScanner
from .jks_scanner import JksScanner
from .lang_scanner import LangScanner
from .mxfile_scanner import MxfileScanner
from .pdf_scanner import PdfScanner
from .pkcs12_scanner import Pkcs12Scanner
from .pptx_scanner import PptxScanner
from .tar_scanner import TarScanner
from .xlsx_scanner import XlsxScanner
from .xml_scanner import XmlScanner
from .zip_scanner import ZipScanner
from ..common.constants import DEFAULT_ENCODING
from ..file_handler.file_path_extractor import FilePathExtractor
from ..file_handler.struct_content_provider import StructContentProvider

logger = logging.getLogger(__name__)


class DeepScanner(
    ByteScanner,  #
    Bzip2Scanner,  #
    DocxScanner,  #
    EncoderScanner,  #
    GzipScanner,  #
    HtmlScanner,  #
    JksScanner,  #
    LangScanner,  #
    PdfScanner,  #
    Pkcs12Scanner,  #
    PptxScanner,  #
    TarScanner,  #
    XmlScanner,  #
    XlsxScanner,  #
    ZipScanner
):  # yapf: disable
    """Advanced scanner with recursive exploring of data"""

    def __init__(self, config: Config, scanner: Scanner) -> None:
        """Initialize Advanced credential scanner.

        Args:
            scanner: CredSweeper scanner object
            config: dictionary variable, stores analyzer features
        """
        self.__config = config
        self.__scanner = scanner

    @property
    def config(self) -> Config:
        return self.__config

    @property
    def scanner(self) -> Scanner:
        return self.__scanner

    @staticmethod
    def get_deep_scanners(data: bytes, file_type: str, depth: int) -> List[Any]:
        """Returns possibly scan methods for the data depends on content"""
        deep_scanners: List[Any] = []
        if Util.is_zip(data):
            if 0 < depth:
                deep_scanners.append(ZipScanner)
            # probably, there might be a docx, xlxs and so on.
            # It might be scanned with text representation in third-party libraries.
            deep_scanners.append(XlsxScanner)
            deep_scanners.append(DocxScanner)
            deep_scanners.append(PptxScanner)
        elif Util.is_bzip2(data):
            if 0 < depth:
                deep_scanners.append(Bzip2Scanner)
        elif Util.is_tar(data):
            if 0 < depth:
                deep_scanners.append(TarScanner)
        elif Util.is_gzip(data):
            if 0 < depth:
                deep_scanners.append(GzipScanner)
        elif Util.is_pdf(data):
            deep_scanners.append(PdfScanner)
        elif Util.is_jks(data):
            deep_scanners.append(JksScanner)
        elif Util.is_asn1(data):
            deep_scanners.append(Pkcs12Scanner)
        elif file_type in [".eml", ".mht"]:
            if Util.is_eml(data):
                deep_scanners.append(EmlScanner)
            elif Util.is_xml(data) and Util.is_html(data):
                deep_scanners.append(HtmlScanner)
            else:
                deep_scanners.append(ByteScanner)
        elif Util.is_xml(data):
            if Util.is_html(data):
                deep_scanners.append(HtmlScanner)
            elif Util.is_mxfile(data):
                deep_scanners.append(MxfileScanner)
            deep_scanners.append(XmlScanner)
        else:
            if 0 < depth:
                deep_scanners.append(EncoderScanner)
                deep_scanners.append(LangScanner)
            deep_scanners.append(ByteScanner)
        return deep_scanners

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def scan(self,
             content_provider: ContentProvider,
             depth: int,
             recursive_limit_size: Optional[int] = None) -> List[Candidate]:
        """Initial scan method to launch recursive scan. Skips ByteScanner to prevent extra scan

            Args:
                content_provider: ContentProvider that might contain raw data
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        recursive_limit_size = recursive_limit_size if isinstance(recursive_limit_size,
                                                                  int) else RECURSIVE_SCAN_LIMITATION
        candidates: List[Candidate] = []
        data: Optional[bytes] = None
        if isinstance(content_provider, TextContentProvider) or isinstance(content_provider, ByteContentProvider):
            # Feature to scan files which might be containers
            data = content_provider.data
            info = "FILE"
        elif isinstance(content_provider, DiffContentProvider) and content_provider.diff:
            candidates = self.scanner.scan(content_provider)
            # Feature to scan binary diffs
            diff = content_provider.diff[0].get("line")
            # the check for legal fix mypy issue
            if isinstance(diff, bytes):
                data = diff
            info = "DIFF"
        else:
            logger.warning(f"Content provider {type(content_provider)} does not support deep scan")
            info = "NA"

        if data:
            data_provider = DataContentProvider(data=data,
                                                file_path=content_provider.file_path,
                                                file_type=content_provider.file_type,
                                                info=content_provider.info or info)
            # iterate for all possibly scanner methods WITHOUT ByteContentProvider for TextContentProvider
            scanner_classes = self.get_deep_scanners(data, content_provider.file_type, depth)
            fallback = True
            for scan_class in scanner_classes:
                if new_candidates := scan_class.data_scan(self, data_provider, depth, recursive_limit_size - len(data)):
                    augment_candidates(candidates, new_candidates)
                    fallback = False
            if fallback and ByteScanner not in scanner_classes and not Util.is_binary(data):
                # wrong assumption case
                fallback_candidates = ByteScanner.data_scan(self, data_provider, depth, recursive_limit_size)
                augment_candidates(candidates, fallback_candidates)
        return candidates

    def recursive_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int = 0,  #
            recursive_limit_size: int = 0) -> List[Candidate]:
        """Recursive function to scan files which might be containers like ZIP archives

            Args:
                data_provider: DataContentProvider object may be a container
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        candidates: List[Candidate] = []
        logger.debug("Start data_scan: size=%d, depth=%d, limit=%d, path=%s, info=%s", len(data_provider.data), depth,
                     recursive_limit_size, data_provider.file_path, data_provider.info)

        if 0 > depth:
            # break recursion if maximal depth is reached
            logger.debug("bottom reached %s recursive_limit_size:%d", data_provider.file_path, recursive_limit_size)
            return candidates

        depth -= 1

        if FilePathExtractor.is_find_by_ext_file(self.config, data_provider.file_type):
            # Skip scanning file and makes fake candidate due the extension is suspicious
            dummy_candidate = Candidate.get_dummy_candidate(self.config, data_provider.file_path,
                                                            data_provider.file_type, data_provider.info,
                                                            FilePathExtractor.FIND_BY_EXT_RULE)
            candidates.append(dummy_candidate)
        else:
            fallback = True
            # iterate for all possibly scanner methods
            scanner_classes = self.get_deep_scanners(data_provider.data, data_provider.file_type, depth)
            for scanner_class in scanner_classes:
                if new_candidates := scanner_class.data_scan(self, data_provider, depth, recursive_limit_size):
                    augment_candidates(candidates, new_candidates)
                    fallback = False
            if fallback and ByteScanner not in scanner_classes and not Util.is_binary(data_provider.data):
                bypass_candidates = ByteScanner.data_scan(self, data_provider, depth, recursive_limit_size)
                augment_candidates(candidates, bypass_candidates)

        return candidates

    def structure_scan(
            self,  #
            struct_provider: StructContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> List[Candidate]:
        """Recursive function to scan structured data

            Args:
                struct_provider: DataContentProvider object may be a container
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        candidates: List[Candidate] = []
        logger.debug("Start struct_scan: depth=%d, limit=%d, path=%s, info=%s", depth, recursive_limit_size,
                     struct_provider.file_path, struct_provider.info)

        if 0 > depth:
            # break recursion if maximal depth is reached
            logger.debug("bottom reached %s recursive_limit_size:%d", struct_provider.file_path, recursive_limit_size)
            return candidates

        depth -= 1

        items: List[Tuple[Union[int, str], Any]] = []
        struct_key: Optional[str] = None
        struct_value: Optional[str] = None
        line_for_keyword_rules = ""
        if isinstance(struct_provider.struct, dict):
            for key, value in struct_provider.struct.items():
                if isinstance(value, (list, tuple)) and 1 == len(value):
                    # simplify some structures like YAML when single item in new line is a value
                    items.append((key, value[0]))
                else:
                    items.append((key, value))
            # for transformation {"key": "api_key", "value": "XXXXXXX"} -> {"api_key": "XXXXXXX"}
            struct_key = struct_provider.struct.get("key")
            struct_value = struct_provider.struct.get("value")
        elif isinstance(struct_provider.struct, list) or isinstance(struct_provider.struct, tuple):
            items = list(enumerate(struct_provider.struct))
        else:
            logger.error("Not supported type:%s val:%s", str(type(struct_provider.struct)), str(struct_provider.struct))

        for key, value in items:
            if isinstance(value, dict) or isinstance(value, (list, tuple)) and 1 < len(value):
                val_struct_provider = StructContentProvider(struct=value,
                                                            file_path=struct_provider.file_path,
                                                            file_type=struct_provider.file_type,
                                                            info=f"{struct_provider.info}|STRUCT:{key}")
                new_candidates = self.structure_scan(val_struct_provider, depth, recursive_limit_size)
                candidates.extend(new_candidates)

            elif isinstance(value, bytes):
                bytes_struct_provider = DataContentProvider(data=value,
                                                            file_path=struct_provider.file_path,
                                                            file_type=struct_provider.file_type,
                                                            info=f"{struct_provider.info}|BYTES:{key}")
                new_limit = recursive_limit_size - len(value)
                new_candidates = self.recursive_scan(bytes_struct_provider, depth, new_limit)
                candidates.extend(new_candidates)

            elif isinstance(value, str):
                data = value.encode(encoding=DEFAULT_ENCODING, errors='replace')
                str_struct_provider = DataContentProvider(data=data,
                                                          file_path=struct_provider.file_path,
                                                          file_type=struct_provider.file_type,
                                                          info=f"{struct_provider.info}|STRING:{key}")
                new_limit = recursive_limit_size - len(str_struct_provider.data)
                new_candidates = self.recursive_scan(str_struct_provider, depth, new_limit)
                candidates.extend(new_candidates)

                # use key = "value" scan for common cases like in TOML
                if isinstance(key, str) and self.scanner.keywords_required_substrings_check(key):
                    line_for_keyword_rules += f"{key} = \"{value}\"; "

            elif isinstance(value, (int, float, datetime.date, datetime.datetime)):
                # use the fields only in case of matched keywords
                if isinstance(key, str) and self.scanner.keywords_required_substrings_check(key):
                    line_for_keyword_rules += f"{key} = \"{value}\"; "

            else:
                logger.warning("Not supported type:%s value(%s)", str(type(value)), str(value))

        if line_for_keyword_rules:
            str_provider = StringContentProvider([line_for_keyword_rules],
                                                 file_path=struct_provider.file_path,
                                                 file_type=".toml",
                                                 info=f"{struct_provider.info}|KEYWORD:`{line_for_keyword_rules}`")
            new_candidates = self.scanner.scan(str_provider)
            augment_candidates(candidates, new_candidates)

        # last check when dictionary is {"key": "api_key", "value": "XXXXXXX"} -> {"api_key": "XXXXXXX"}
        if isinstance(struct_key, str) and isinstance(struct_value, str):
            line_for_keyword_rules = f"{struct_key} = \"{struct_value}\""
            key_value_provider = StringContentProvider(
                [line_for_keyword_rules],
                file_path=struct_provider.file_path,
                file_type=".toml",
                info=f"{struct_provider.info}|KEY_VALUE:`{line_for_keyword_rules}`")
            new_candidates = self.scanner.scan(key_value_provider)
            augment_candidates(candidates, new_candidates)
        return candidates



import io
import logging
from abc import ABC
from typing import List, Optional

import pandas as pd

from credsweeper.credentials import Candidate
from credsweeper.credentials.augment_candidates import augment_candidates
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class XlsxScanner(AbstractScanner, ABC):
    """Implements xlsx scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan xlsx text elements for all slides"""
        try:
            candidates = []
            book = pd.read_excel(io.BytesIO(data_provider.data), sheet_name=None, header=None)
            for sheet_name, sheet_data in book.items():
                sheet_info = f"{data_provider.info}|{sheet_name}"
                # replace open xml carriage returns _x000D_ before line feed only
                df = sheet_data.replace(to_replace="_x000D_\n", value='\n', regex=True).fillna('').astype(str)
                for row_pos, row in enumerate(df.values):
                    for col_pos, cell in enumerate(row):
                        cell_info = f"{sheet_info}:{Util.get_excel_column_name(col_pos)}{row_pos + 1}"
                        cell_provider = StringContentProvider(lines=cell.splitlines(),
                                                              file_path=data_provider.file_path,
                                                              file_type=data_provider.file_type,
                                                              info=cell_info)
                        cell_candidates = self.scanner.scan(cell_provider)
                        candidates.extend(cell_candidates)
                    row_line = '\t'.join(row)
                    row_provider = StringContentProvider(lines=[row_line],
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{sheet_info}:R{row_pos + 1}")
                    row_candidates = self.scanner.scan(row_provider)
                    augment_candidates(candidates, row_candidates)

            return candidates
        except Exception as xlsx_exc:
            logger.error(f"{data_provider.file_path}:{xlsx_exc}")
        return None



import math

from credsweeper.common.constants import Chars
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueEntropyBase36Check(Filter):
    """Check that candidate have Shanon Entropy (for [a-z0-9])"""

    def __init__(self, config: Config = None) -> None:
        pass

    @staticmethod
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal entropy for size of random data. Precalculated data is applied for speedup"""
        if 15 == x:
            y = 3.43
        elif 24 == x:
            y = 3.91
        elif 25 == x:
            y = 3.95
        elif 10 <= x:
            # approximation does not exceed standard deviation
            y = 0.7 * math.log2(x) + 0.7
        else:
            y = 0
        return y

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        entropy = Util.get_shannon_entropy(line_data.value, Chars.BASE36_CHARS.value)
        min_entropy = ValueEntropyBase36Check.get_min_data_entropy(len(line_data.value))
        return min_entropy > entropy or 0 == min_entropy



from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueBlocklistCheck(Filter):
    """Check that words from block list is lest that 70% of candidate value length."""

    NOT_ALLOWED = [
        "true",
        "false",
        "null",
        "bearer",
        "string",
    ]

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """

        value = line_data.value.lower()
        for not_allowed in self.NOT_ALLOWED:
            if not_allowed in value and len(not_allowed) / len(value) >= 0.7:
                return True

        return False



import logging
from abc import ABC
from typing import List, Optional

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider

logger = logging.getLogger(__name__)


class LangScanner(AbstractScanner, ABC):
    """Implements scanning of data if it is a script of some markup language"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to represent data as markup language and scan as structure"""
        if data_provider.represent_as_structure():
            struct_data_provider = StructContentProvider(struct=data_provider.structure,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|STRUCT")
            return self.structure_scan(struct_data_provider, depth, recursive_limit_size)
        return None



from credsweeper.file_handler.byte_content_provider import ByteContentProvider
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.file_handler.text_content_provider import TextContentProvider

__all__ = [
    'ByteContentProvider',  #
    'ContentProvider',  #
    'DataContentProvider',  #
    'DiffContentProvider',  #
    'StringContentProvider',  #
    'TextContentProvider',  #
]



import io
import logging
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Union, Tuple, Generator

from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class TextContentProvider(ContentProvider):
    """Provide access to analysis targets for full-text file scanning.

    Parameters:
        file_path: string, path to file

    """

    def __init__(self,
                 file_path: Union[str, Path, Tuple[Union[str, Path], io.BytesIO]],
                 file_type: Optional[str] = None,
                 info: Optional[str] = None) -> None:
        _path = str(file_path[0]) if isinstance(file_path, tuple) else str(file_path)
        self.__io = file_path[1] if isinstance(file_path, tuple) else None
        self.__data: Optional[bytes] = None
        self.__lines: Optional[List[str]] = None
        super().__init__(file_path=_path, file_type=file_type, info=info)

    @cached_property
    def data(self) -> Optional[bytes]:
        """data RO getter for TextContentProvider"""
        if self.__data is None:
            if isinstance(self.__io, io.BytesIO) and self.__io:
                self.__data = self.__io.read()
            else:
                self.__data = Util.read_data(self.file_path)
        return self.__data

    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        self.__data = None
        if hasattr(self, "data"):
            delattr(self, "data")
        self.__lines = None
        if hasattr(self, "lines"):
            delattr(self, "lines")
        if isinstance(self.__io, io.BytesIO) and self.__io and not self.__io.closed:
            self.__io.close()

    @cached_property
    def lines(self) -> Optional[List[str]]:
        """lines getter for TextContentProvider"""
        if self.__lines is None:
            self.__lines = Util.decode_bytes(self.data)
        return self.__lines if self.__lines is not None else []

    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Load and preprocess file content to scan.

        Args:
            min_len: minimal line length to scan

        Return:
            list of analysis targets based on every row in file

        """
        lines: Optional[List[str]] = None
        line_nums: Optional[List[int]] = None

        if Util.get_extension(self.file_path) == ".xml":
            try:
                # append line ending for correct xml line numeration
                xml_lines = [f"{line}\n" for line in self.lines]
                lines, line_nums = Util.get_xml_from_lines(xml_lines)
            except Exception as exc:
                logger.error(f"Cannot parse to xml {exc}")

        if lines is None:
            lines = self.lines

        return self.lines_to_targets(min_len, lines, line_nums)
